@@ -24,6 +24,7 @@
  * The current implementation remembers all file names ever used.
  */
 
+
 #include "vim.h"
 
 static void	enter_buffer __ARGS((BUF *));
@@ -92,12 +93,6 @@ open_buffer(read_stdin)
     else if (read_stdin)
 	retval = readfile(NULL, NULL, (linenr_t)0,
 		       (linenr_t)0, (linenr_t)MAXLNUM, READ_NEW + READ_STDIN);
-    else
-    {
-	MSG("Empty Buffer");
-	msg_col = 0;
-	msg_didout = FALSE;	/* overwrite this message whenever you like */
-    }
 
     /* if first time loading this buffer, init chartab */
     if (curbuf->b_flags & BF_NEVERLOADED)
@@ -125,7 +120,7 @@ open_buffer(read_stdin)
 	curbuf->b_notedited = TRUE;
 
 #ifdef AUTOCMD
-    apply_autocmds(EVENT_BUFENTER, NULL, NULL, FALSE);
+    apply_autocmds(EVENT_BUFENTER, NULL, NULL, FALSE, curbuf);
 #endif
 
     if (retval != FAIL)
@@ -180,17 +175,15 @@ close_buffer(win, buf, free_buf, del_buf)
     int	    free_buf;
     int	    del_buf;
 {
-    /* always remove Help buffers */
-    if (buf->b_help)
-    {
-	free_buf = TRUE;
-	del_buf = TRUE;
-    } 
-
     if (buf->b_nwindows > 0)
 	--buf->b_nwindows;
     if (buf->b_nwindows == 0 && win != NULL)
+    {
 	set_last_cursor(win);	/* may set b_last_cursor */
+	if (win == curwin && curwin->w_cursor.lnum != 1)
+				/* and remember last cursor position */
+	    buflist_setfpos(buf, curwin->w_cursor.lnum, curwin->w_cursor.col);
+    }
     if (buf->b_nwindows > 0 || !free_buf)
     {
 	if (buf == curbuf)
@@ -257,12 +250,12 @@ buf_freeall(buf, del_buf)
     int	    del_buf;	/* buffer is going to be deleted */
 {
 #ifdef AUTOCMD
-    apply_autocmds(EVENT_BUFUNLOAD, buf->b_fname, buf->b_fname, FALSE);
+    apply_autocmds(EVENT_BUFUNLOAD, buf->b_fname, buf->b_fname, FALSE, buf);
     if (!buf_valid(buf))	    /* autocommands may delete the buffer */
 	return;
     if (del_buf)
     {
-	apply_autocmds(EVENT_BUFDELETE, buf->b_fname, buf->b_fname, FALSE);
+	apply_autocmds(EVENT_BUFDELETE, buf->b_fname, buf->b_fname, FALSE, buf);
 	if (!buf_valid(buf))	    /* autocommands may delete the buffer */
 	    return;
     }
@@ -338,7 +331,17 @@ do_bufdel(command, arg, addr_count, start_bnr, end_bnr, forceit)
     char_u  *p;
 
     if (addr_count == 0)
+    {
+#ifdef USE_GUI_GTK
+	int bufnum = curbuf->b_fnum;
+
+	/* remove entry from buflist window */
+	if (do_buffer(command, DOBUF_CURRENT, FORWARD, 0, forceit) == OK)
+	    buflist_delete_entry(bufnum);
+#else
 	(void)do_buffer(command, DOBUF_CURRENT, FORWARD, 0, forceit);
+#endif
+    }
     else
     {
 	if (addr_count == 2)
@@ -481,6 +484,10 @@ do_buffer(action, start, dir, count, forceit)
 		if (buf == NULL)
 		    buf = lastbuf;
 	    }
+	    /* in non-help buffer, skip help buffers, and vv */
+	    if ((buf->b_help) !=
+			     (start == DOBUF_LAST ? lastbuf : curbuf)->b_help)
+		 count++;
 	}
     }
 
@@ -499,6 +506,10 @@ do_buffer(action, start, dir, count, forceit)
 	return FAIL;
     }
 
+#ifdef USE_GUI
+    need_mouse_correct = TRUE;
+#endif
+
     /*
      * delete buffer buf from memory and/or the list
      */
@@ -576,7 +587,8 @@ do_buffer(action, start, dir, count, forceit)
 		forward = FALSE;
 		continue;
 	    }
-	    if (buf->b_ml.ml_mfp != NULL)
+	    /* in non-help buffer, try to skip help buffers, and vv */
+	    if (buf->b_ml.ml_mfp != NULL && buf->b_help == curbuf->b_help)
 		break;
 	    if (forward)
 		buf = buf->b_next;
@@ -622,7 +634,7 @@ do_buffer(action, start, dir, count, forceit)
     delbuf = curbuf;
 
 #ifdef AUTOCMD
-    apply_autocmds(EVENT_BUFLEAVE, NULL, NULL, FALSE);
+    apply_autocmds(EVENT_BUFLEAVE, NULL, NULL, FALSE, curbuf);
     if (buf_valid(delbuf))
 #endif
     {
@@ -658,9 +670,9 @@ enter_buffer(buf)
     else
     {
 	need_fileinfo = TRUE;		/* display file info after redraw */
-	buf_check_timestamp(curbuf);	/* check if file has changed */
+	(void)buf_check_timestamp(curbuf);	/* check if file has changed */
 #ifdef AUTOCMD
-	apply_autocmds(EVENT_BUFENTER, NULL, NULL, FALSE);
+	apply_autocmds(EVENT_BUFENTER, NULL, NULL, FALSE, curbuf);
 #endif
     }
     buflist_getfpos();			/* restore curpos.lnum and possibly
@@ -718,7 +730,13 @@ buflist_new(ffname, sfname, lnum, use_curbuf)
     if (use_curbuf && curbuf != NULL && curbuf->b_ffname == NULL &&
 		curbuf->b_nwindows <= 1 &&
 		(curbuf->b_ml.ml_mfp == NULL || bufempty()))
+    {
 	buf = curbuf;
+#ifdef AUTOCMD
+	/* It's like this buffer is deleted. */
+	apply_autocmds(EVENT_BUFDELETE, NULL, NULL, FALSE, curbuf);
+#endif
+    }
     else
     {
 	buf = (BUF *)alloc_clear((unsigned)sizeof(BUF));
@@ -823,11 +841,29 @@ free_buf_options(buf, free_p_ff)
     int	    free_p_ff;	    /* also free 'fileformat'? */
 {
     if (free_p_ff)
+    {
 	free_string_option(buf->b_p_ff);
+#ifdef MULTI_BYTE
+	free_string_option(buf->b_p_fe);
+#endif
+    }
+#ifdef CRYPTV
+    free_string_option(buf->b_p_key);
+#endif
+    free_string_option(buf->b_p_mps);
     free_string_option(buf->b_p_fo);
     free_string_option(buf->b_p_isk);
     free_string_option(buf->b_p_com);
     free_string_option(buf->b_p_nf);
+#ifdef SYNTAX_HL
+    free_string_option(buf->b_p_syn);
+#endif
+#ifdef AUTOCMD
+    free_string_option(buf->b_p_ft);
+#endif
+#ifdef WANT_FILETYPE
+    free_string_option(buf->b_p_oft);
+#endif
 #ifdef CINDENT
     free_string_option(buf->b_p_cink);
     free_string_option(buf->b_p_cino);
@@ -843,7 +879,7 @@ free_buf_options(buf, free_p_ff)
 /*
  * get alternate file n
  * set linenr to lnum or altfpos.lnum if lnum == 0
- * 	also set cursor column to altfpos.col if 'startofline' is not set.
+ *	also set cursor column to altfpos.col if 'startofline' is not set.
  * if (options & GETF_SETMARK) call setpcmark()
  * if (options & GETF_ALT) we are jumping to an alternate file.
  *
@@ -1017,6 +1053,8 @@ buflist_findpat(pattern, pattern_end)
     return fnum;
 }
 
+#ifdef CMDLINE_COMPL
+
 /*
  * Find all buffer names that match.
  * For command line expansion of ":buf" and ":sbuf".
@@ -1100,6 +1138,8 @@ ExpandBufnames(pat, num_file, file, options)
     return (count == 0 ? FAIL : OK);
 }
 
+#endif /* CMDLINE_COMPL */
+
 /*
  * Check for a match on the file name for buffer "buf" with regex prog "prog".
  */
@@ -1233,7 +1273,7 @@ buflist_setfpos(buf, lnum, col)
 
 /*
  * find the position (lnum and col) for the buffer 'buf' for the current window
- * returns a pointer to no_position if no position is found 
+ * returns a pointer to no_position if no position is found
  */
     static FPOS *
 buflist_findfpos(buf)
@@ -1274,13 +1314,21 @@ buflist_list()
     BUF		*buf;
     int		len;
 
+#ifdef USE_GUI_GTK
+    if (browse)
+    {
+	gtk_buffer_list();
+	return;
+    }
+#endif
+
     for (buf = firstbuf; buf != NULL && !got_int; buf = buf->b_next)
     {
 	msg_putchar('\n');
 	if (buf->b_fname == NULL)
 	    STRCPY(NameBuff, "No File");
 	else
-	    home_replace(buf, buf->b_fname, NameBuff, MAXPATHL);
+	    home_replace(buf, buf->b_fname, NameBuff, MAXPATHL, TRUE);
 
 	sprintf((char *)IObuff, "%3d %c%c%c \"",
 		buf->b_fnum,
@@ -1311,6 +1359,7 @@ buflist_list()
     }
 }
 
+
 /*
  * Get file name and line number for file 'fnum'.
  * Used by DoOneCmd() for translating '%' and '#'.
@@ -1435,17 +1484,19 @@ setaltfname(ffname, sfname, lnum)
 
 /*
  * Get alternate file name for current window.
- * Return NULL if there isn't any, and give error message.
+ * Return NULL if there isn't any, and give error message if requested.
  */
     char_u  *
-getaltfname()
+getaltfname(errmsg)
+    int		errmsg;		/* give error message */
 {
     char_u	*fname;
     linenr_t	dummy;
 
     if (buflist_name_nr(0, &fname, &dummy) == FAIL)
     {
-	emsg(e_noalt);
+	if (errmsg)
+	    emsg(e_noalt);
 	return NULL;
     }
     return fname;
@@ -1538,7 +1589,7 @@ fileinfo(fullname, shorthelp, dont_truncate)
 	else
 	    name = curbuf->b_ffname;
 	home_replace(shorthelp ? curbuf : NULL, name, p,
-						(int)(IOSIZE - (p - buffer)));
+					  (int)(IOSIZE - (p - buffer)), TRUE);
     }
 
     sprintf((char *)buffer + STRLEN(buffer),
@@ -1581,7 +1632,7 @@ fileinfo(fullname, shorthelp, dont_truncate)
     if (dont_truncate)
 	msg(buffer);
     else
-	msg_trunc(buffer);
+	msg_trunc_attr(buffer, FALSE, 0);
 
     vim_free(buffer);
 }
@@ -1675,7 +1726,7 @@ maketitle()
 	if (buf == NULL)
 	    return;
 	STRCPY(buf, "VIM - ");
-	home_replace(curbuf, curbuf->b_ffname, buf + 6, IOSIZE - 6);
+	home_replace(curbuf, curbuf->b_ffname, buf + 6, IOSIZE - 6, TRUE);
 	append_arg_number(buf, FALSE, IOSIZE);
 	if (p_titlelen > 0)
 	{
@@ -1685,7 +1736,7 @@ maketitle()
 	    len = STRLEN(buf);
 	    if (len > maxlen)
 	    {
-		vim_memmove(buf + 6, buf + 6 + len - maxlen,
+		mch_memmove(buf + 6, buf + 6 + len - maxlen,
 							  (size_t)maxlen - 5);
 		buf[5] = '<';
 	    }
@@ -1694,33 +1745,36 @@ maketitle()
 	i_name = gettail(curbuf->b_ffname); /* use file name only for icon */
     }
 
-    if (*p_titlestring)
-	t_name = p_titlestring;
-    if (*p_iconstring)
-	i_name = p_iconstring;
-
     vim_free(lasttitle);
-    if (p_title && (lasttitle = alloc((unsigned)(vim_strsize(t_name) + 1)))
-								      != NULL)
+    lasttitle = NULL;
+    if (p_title)
     {
-	*lasttitle = NUL;
-	while (*t_name)
-	    STRCAT(lasttitle, transchar(*t_name++));
+	if (*p_titlestring)
+	    lasttitle = vim_strsave(p_titlestring);
+	else if ((lasttitle = alloc((unsigned)(vim_strsize(t_name) + 1)))
+								      != NULL)
+	{
+	    *lasttitle = NUL;
+	    while (*t_name)
+		STRCAT(lasttitle, transchar(*t_name++));
+	}
     }
-    else
-	lasttitle = NULL;
     vim_free(buf);
 
     vim_free(lasticon);
-    if (p_icon && (lasticon = alloc((unsigned)(vim_strsize(i_name) + 1)))
-								      != NULL)
+    lasticon = NULL;
+    if (p_icon)
     {
-	*lasticon = NUL;
-	while (*i_name)
-	    STRCAT(lasticon, transchar(*i_name++));
+	if (*p_iconstring)
+	    lasticon = vim_strsave(p_iconstring);
+	else if ((lasticon = alloc((unsigned)(vim_strsize(i_name) + 1)))
+								      != NULL)
+	{
+	    *lasticon = NUL;
+	    while (*i_name)
+		STRCAT(lasticon, transchar(*i_name++));
+	}
     }
-    else
-	lasticon = NULL;
 
     resettitle();
 }
@@ -1844,6 +1898,10 @@ do_arg_all(count, forceit)
     if (opened == NULL)
 	return;
 
+#ifdef USE_GUI
+    need_mouse_correct = TRUE;
+#endif
+
     /*
      * Try closing all windows that are not in the argument list.
      * When 'hidden' or "forceit" set the buffer becomes hidden.
@@ -1998,6 +2056,10 @@ do_buffer_all(count, all)
 
     setpcmark();
 
+#ifdef USE_GUI
+    need_mouse_correct = TRUE;
+#endif
+
     /*
      * Close superfluous windows (two windows for the same buffer).
      */
@@ -2112,10 +2174,17 @@ do_modelines()
 {
     linenr_t	    lnum;
     int		    nmlines;
+    static int	    entered = 0;
 
     if (!curbuf->b_p_ml || (nmlines = (int)p_mls) == 0)
 	return;
 
+    /* Disallow recursive entry here.  Can happen when executing a modeline
+     * triggers an autocommand, which reloads modelines with a ":do". */
+    if (entered)
+	return;
+
+    ++entered;
     for (lnum = 1; lnum <= curbuf->b_ml.ml_line_count && lnum <= nmlines;
 								       ++lnum)
 	if (chk_modeline(lnum) == FAIL)
@@ -2125,6 +2194,7 @@ do_modelines()
 			  lnum > curbuf->b_ml.ml_line_count - nmlines; --lnum)
 	if (chk_modeline(lnum) == FAIL)
 	    nmlines = 0;
+    --entered;
 }
 
 /*
@@ -2200,7 +2270,7 @@ chk_modeline(lnum)
 	    }
 
 	    *e = NUL;			/* truncate the set command */
-	    if (do_set(s) == FAIL)	/* stop if error found */
+	    if (do_set(s, TRUE) == FAIL)/* stop if error found */
 	    {
 		retval = FAIL;
 		break;
@@ -2283,7 +2353,7 @@ write_viminfo_bufferlist(fp)
 	if (buf->b_fname == NULL || buf->b_help || removable(buf->b_ffname))
 	    continue;
 
-	home_replace(NULL, buf->b_ffname, NameBuff, MAXPATHL);
+	home_replace(NULL, buf->b_ffname, NameBuff, MAXPATHL, TRUE);
 
 	fprintf(fp, "%%%s\t%ld\t%d\n", NameBuff,
 			(long)buf->b_last_cursor.lnum,
