@@ -42,8 +42,10 @@ struct buffheader
 static struct buffheader stuffbuff = {{NULL, {NUL}}, NULL, 0, 0};
 static struct buffheader redobuff = {{NULL, {NUL}}, NULL, 0, 0};
 static struct buffheader old_redobuff = {{NULL, {NUL}}, NULL, 0, 0};
+#if defined(AUTOCMD) || defined(WANT_EVAL) || defined(PROTO)
 static struct buffheader save_redobuff = {{NULL, {NUL}}, NULL, 0, 0};
 static struct buffheader save_old_redobuff = {{NULL, {NUL}}, NULL, 0, 0};
+#endif
 static struct buffheader recordbuff = {{NULL, {NUL}}, NULL, 0, 0};
 
 /*
@@ -192,13 +194,25 @@ get_recorded()
 
     p = get_bufcont(&recordbuff, TRUE);
     free_buff(&recordbuff);
+
     /*
      * Remove the characters that were added the last time, these must be the
      * (possibly mapped) characters that stopped recording.
      */
     len = STRLEN(p);
     if ((int)len >= last_recorded_len)
-	p[len - last_recorded_len] = NUL;
+    {
+	len -= last_recorded_len;
+	p[len] = NUL;
+    }
+
+    /*
+     * When stopping recording from Insert mode with CTRL-O q, also remove the
+     * CTRL-O.
+     */
+    if (len > 0 && restart_edit && p[len - 1] == Ctrl('O'))
+	p[len - 1] = NUL;
+
     return (p);
 }
 
@@ -405,31 +419,41 @@ ResetRedobuff()
     }
 }
 
+#if defined(AUTOCMD) || defined(WANT_EVAL) || defined(PROTO)
 /*
  * Save redobuff and old_redobuff to save_redobuff and save_old_redobuff.
- * Used before executing autocommands.
+ * Used before executing autocommands and user functions.
  */
+static int save_level = 0;
+
     void
 saveRedobuff()
 {
-    save_redobuff = redobuff;
-    redobuff.bh_first.b_next = NULL;
-    save_old_redobuff = old_redobuff;
-    old_redobuff.bh_first.b_next = NULL;
+    if (save_level++ == 0)
+    {
+	save_redobuff = redobuff;
+	redobuff.bh_first.b_next = NULL;
+	save_old_redobuff = old_redobuff;
+	old_redobuff.bh_first.b_next = NULL;
+    }
 }
 
 /*
  * Restore redobuff and old_redobuff from save_redobuff and save_old_redobuff.
- * Used after executing autocommands.
+ * Used after executing autocommands and user functions.
  */
     void
 restoreRedobuff()
 {
-    free_buff(&redobuff);
-    redobuff = save_redobuff;
-    free_buff(&old_redobuff);
-    old_redobuff = save_old_redobuff;
+    if (--save_level == 0)
+    {
+	free_buff(&redobuff);
+	redobuff = save_redobuff;
+	free_buff(&old_redobuff);
+	old_redobuff = save_old_redobuff;
+    }
 }
+#endif
 
     void
 AppendToRedobuff(s)
@@ -572,6 +596,7 @@ start_redo(count, old_redo)
     {
 	VIsual = curwin->w_cursor;
 	VIsual_active = TRUE;
+	VIsual_reselect = TRUE;
 	redo_VIsual_busy = TRUE;
 	c = read_redo(FALSE, old_redo);
     }
@@ -686,7 +711,7 @@ ins_typebuf(str, noremap, offset, nottyped)
     if (offset == 0 && addlen <= typeoff)
     {
 	typeoff -= addlen;
-	vim_memmove(typebuf + typeoff, str, (size_t)addlen);
+	mch_memmove(typebuf + typeoff, str, (size_t)addlen);
     }
     /*
      * Need to allocate new buffer.
@@ -715,19 +740,19 @@ ins_typebuf(str, noremap, offset, nottyped)
 	typebuflen = newlen;
 
 	/* copy the old chars, before the insertion point */
-	vim_memmove(s1 + newoff, typebuf + typeoff, (size_t)offset);
+	mch_memmove(s1 + newoff, typebuf + typeoff, (size_t)offset);
 	/* copy the new chars */
-	vim_memmove(s1 + newoff + offset, str, (size_t)addlen);
+	mch_memmove(s1 + newoff + offset, str, (size_t)addlen);
 	/* copy the old chars, after the insertion point, including the	NUL at
 	 * the end */
-	vim_memmove(s1 + newoff + offset + addlen, typebuf + typeoff + offset,
+	mch_memmove(s1 + newoff + offset + addlen, typebuf + typeoff + offset,
 					      (size_t)(typelen - offset + 1));
 	if (typebuf != typebuf_init)
 	    vim_free(typebuf);
 	typebuf = s1;
 
-	vim_memmove(s2 + newoff, noremapbuf + typeoff, (size_t)offset);
-	vim_memmove(s2 + newoff + offset + addlen,
+	mch_memmove(s2 + newoff, noremapbuf + typeoff, (size_t)offset);
+	mch_memmove(s2 + newoff + offset + addlen,
 		   noremapbuf + typeoff + offset, (size_t)(typelen - offset));
 	if (noremapbuf != noremapbuf_init)
 	    vim_free(noremapbuf);
@@ -768,6 +793,15 @@ typebuf_typed()
     return typemaplen == 0;
 }
 
+/*
+ * Return the number of characters that are mapped (or not typed).
+ */
+    int
+typebuf_maplen()
+{
+    return typemaplen;
+}
+
 /*
  * remove "len" characters from typebuf[typeoff + offset]
  */
@@ -795,16 +829,16 @@ del_typebuf(len, offset)
 	 */
 	if (typeoff > MAXMAPLEN)
 	{
-	    vim_memmove(typebuf + MAXMAPLEN, typebuf + typeoff, (size_t)offset);
-	    vim_memmove(noremapbuf + MAXMAPLEN, noremapbuf + typeoff,
+	    mch_memmove(typebuf + MAXMAPLEN, typebuf + typeoff, (size_t)offset);
+	    mch_memmove(noremapbuf + MAXMAPLEN, noremapbuf + typeoff,
 							      (size_t)offset);
 	    typeoff = MAXMAPLEN;
 	}
 	/* adjust typebuf (include the NUL at the end) */
-	vim_memmove(typebuf + typeoff + offset, typebuf + i + len,
+	mch_memmove(typebuf + typeoff + offset, typebuf + i + len,
 					      (size_t)(typelen - offset + 1));
 	/* adjust noremapbuf[] */
-	vim_memmove(noremapbuf + typeoff + offset, noremapbuf + i + len,
+	mch_memmove(noremapbuf + typeoff + offset, noremapbuf + i + len,
 						  (size_t)(typelen - offset));
     }
 
@@ -1049,6 +1083,27 @@ vgetc()
 		continue;
 	    }
 	    c = TO_SPECIAL(c2, c);
+
+#ifdef USE_GUI_WIN32
+	    /* Handle K_TEAROFF here, the caller of vgetc() doesn't need to
+	     * know that a menu was torn off */
+	    if (c == K_TEAROFF)
+	    {
+		char_u	name[200];
+		int	i;
+
+		/* get menu path, it ends with a <CR> */
+		for (i = 0; (c = vgetorpeek(TRUE)) != '\r'; )
+		{
+		    name[i] = c;
+		    if (i < 199)
+			++i;
+		}
+		name[i] = NUL;
+		gui_make_tearoff(name);
+		continue;
+	    }
+#endif
 	}
 #ifdef MSDOS
 	/*
@@ -1410,6 +1465,19 @@ vgetorpeek(advance)
 								 local_State);
 				if (idx != MENU_INDEX_INVALID)
 				{
+				    /*
+				     * In Select mode, a Visual mode menu is
+				     * used.  Switch to Visual mode
+				     * temporarily.  Append K_SELECT to switch
+				     * back to Select mode.
+				     */
+				    if (VIsual_active && VIsual_select)
+				    {
+					VIsual_select = FALSE;
+					(void)ins_typebuf(K_SELECT_STRING, -1,
+							  0, TRUE);
+				    }
+
 				    ins_typebuf(current_menu->strings[idx],
 					current_menu->noremap[idx] ? -1 : 0,
 					0, TRUE);
@@ -1448,6 +1516,18 @@ vgetorpeek(advance)
 			    c = -1;
 			    break;
 			}
+
+			/*
+			 * In Select mode, a Visual mode mapping is used.
+			 * Switch to Visual mode temporarily.  Append K_SELECT
+			 * to switch back to Select mode.
+			 */
+			if (VIsual_active && VIsual_select)
+			{
+			    VIsual_select = FALSE;
+			    (void)ins_typebuf(K_SELECT_STRING, -1, 0, TRUE);
+			}
+
 			/*
 			 * Insert the 'to' part in the typebuf.
 			 * If 'from' field is the same as the start of the
@@ -1789,7 +1869,7 @@ inchar(buf, maxlen, wait_time)
 #endif
 	if (buf[0] == NUL || (buf[0] == K_SPECIAL && c < 0))
 	{
-	    vim_memmove(buf + 3, buf + 1, (size_t)i);
+	    mch_memmove(buf + 3, buf + 1, (size_t)i);
 	    buf[2] = K_THIRD(buf[0]);
 	    buf[1] = K_SECOND(buf[0]);
 	    buf[0] = K_SPECIAL;
@@ -1833,15 +1913,16 @@ inchar(buf, maxlen, wait_time)
  * Return 0 for success
  *	  1 for invalid arguments
  *	  2 for no match
- *	  3 for ambiguety
+ *	  3 for ambiguity
  *	  4 for out of mem
  */
     int
-do_map(maptype, keys, mode, abbrev)
+do_map(maptype, keys, mode, abbrev, ambig)
     int	    maptype;
     char_u  *keys;
     int	    mode;
     int	    abbrev;		/* not a mapping but an abbreviation */
+    char_u  **ambig;		/* pointer to mapping that caused ambiguity */
 {
     struct mapblock	*mp, **mpp;
     char_u		*arg;
@@ -1898,9 +1979,14 @@ do_map(maptype, keys, mode, abbrev)
      * replace_termcodes() also removes CTRL-Vs and sometimes backslashes.
      */
     if (haskey)
-	keys = replace_termcodes(keys, &keys_buf, TRUE);
+	keys = replace_termcodes(keys, &keys_buf, TRUE, TRUE);
     if (hasarg)
-	arg = replace_termcodes(arg, &arg_buf, FALSE);
+    {
+	if (STRICMP(arg, "<nop>") == 0)	    /* "<Nop>" means nothing */
+	    arg = (char_u *)"";
+	else
+	    arg = replace_termcodes(arg, &arg_buf, FALSE, TRUE);
+    }
 
 #ifdef FKMAP
     /*
@@ -2030,6 +2116,8 @@ do_map(maptype, keys, mode, abbrev)
 				mpp = &(mp->m_next);
 				continue;
 			    }
+			    if (ambig != NULL)
+				*ambig = p;
 			    retval = 3;
 			    goto theend;
 			}
@@ -2107,6 +2195,11 @@ do_map(maptype, keys, mode, abbrev)
 	goto theend;
     }
 
+#ifdef USE_GUI_WIN32
+    /* If CTRL-C has been mapped, don't use it for Interrupting */
+    if (*keys == Ctrl('C'))
+	mapped_ctrl_c = TRUE;
+#endif
     mp->m_keys = vim_strsave(keys);
     mp->m_str = vim_strsave(arg);
     if (mp->m_keys == NULL || mp->m_str == NULL)
@@ -2311,16 +2404,181 @@ showmap(mp)
 	++len;
     } while (len < 12);
     if (mp->m_noremap)
-	msg_putchar('*');
+	msg_puts_attr((char_u *)"*", hl_attr(HLF_8));
     else
 	msg_putchar(' ');
     /* Use FALSE below if we only want things like <Up> to show up as such on
      * the rhs, and not M-x etc, TRUE gets both -- webb
      */
-    msg_outtrans_special(mp->m_str, TRUE);
+    if (*mp->m_str == NUL)
+	msg_puts_attr((char_u *)"<Nop>", hl_attr(HLF_8));
+    else
+	msg_outtrans_special(mp->m_str, FALSE);
     out_flush();			/* show one line at a time */
 }
 
+#ifdef CMDLINE_COMPL
+/*
+ * Used below when expanding mapping/abbreviation names.
+ */
+static int	expand_mapmodes = 0;
+static int	expand_isabbrev = 0;
+
+/*
+ * Work out what to complete when doing command line completion of mapping
+ * or abbreviation names.
+ */
+    char_u *
+set_context_in_map_cmd(cmd, arg, forceit, isabbrev, isunmap, cmdidx)
+    char_u	*cmd;
+    char_u	*arg;
+    int		forceit;	/* TRUE if '!' given */
+    int		isabbrev;	/* TRUE if abbreviation */
+    int		isunmap;	/* TRUE if unmap/unabbrev command */
+    CMDIDX	cmdidx;
+{
+    if (forceit && cmdidx != CMD_map && cmdidx != CMD_unmap)
+	expand_context = EXPAND_NOTHING;
+    else
+    {
+	if (isunmap)
+	    expand_mapmodes = get_map_mode(&cmd, forceit || isabbrev);
+	else
+	{
+	    expand_mapmodes = INSERT + CMDLINE;
+	    if (!isabbrev)
+		expand_mapmodes += VISUAL + NORMAL + OP_PENDING;
+	}
+	expand_isabbrev = isabbrev;
+	expand_context = EXPAND_MAPPINGS;
+	expand_pattern = arg;
+    }
+
+    return NULL;
+}
+
+#ifdef HAVE_QSORT
+/*
+ * Compare function for qsort() below.
+ */
+static int
+#ifdef __BORLANDC__
+_RTLENTRYF
+#endif
+compare_mkeys __ARGS((const void *s1, const void *s2));
+
+    static int
+#ifdef __BORLANDC__
+_RTLENTRYF
+#endif
+compare_mkeys(s1, s2)
+    const void	*s1;
+    const void	*s2;
+{
+    return STRCMP(*(char **)s1, *(char **)s2);
+}
+#endif
+
+/*
+ * Find all mapping/abbreviation names that match regexp 'prog'.
+ * For command line expansion of ":[un]map" and ":[un]abbrev" in all modes.
+ * Return OK if matches found, FAIL otherwise.
+ */
+    int
+ExpandMappings(prog, num_file, file)
+    vim_regexp	*prog;
+    int		*num_file;
+    char_u	***file;
+{
+    struct mapblock	*mp;
+    int			hash;
+    int			count = 0;
+    int			round;
+    char_u		*p;
+
+    validate_maphash();
+
+    *num_file = 0;		    /* return values in case of FAIL */
+    *file = NULL;
+
+    /*
+     * round == 1: Count the matches.
+     * round == 2: Build the array to keep the matches.
+     */
+    for (round = 1; round <= 2; ++round)
+    {
+	count = 0;
+	for (hash = 0; hash < 256; ++hash)
+	{
+	    if (expand_isabbrev)
+	    {
+		if (hash)	/* only one abbrev list */
+		    break; /* for (hash) */
+		mp = first_abbr;
+	    }
+	    else
+		mp = maphash[hash];
+	    for (; mp; mp = mp->m_next)
+	    {
+		if (mp->m_mode & expand_mapmodes)
+		{
+		    p = translate_mapping(mp->m_keys, TRUE);
+		    if (p != NULL && vim_regexec(prog, p, TRUE))
+		    {
+			if (round == 1)
+			    ++count;
+			else
+			{
+			    (*file)[count++] = p;
+			    p = NULL;
+			}
+		    }
+		    vim_free(p);
+		}
+	    } /* for (mp) */
+	} /* for (hash) */
+
+	if (count == 0)			/* no match found */
+	    break; /* for (round) */
+
+	if (round == 1)
+	{
+	    *file = (char_u **)alloc((unsigned)(count * sizeof(char_u *)));
+	    if (*file == NULL)
+	    {
+		vim_free(prog);
+		return FAIL;
+	    }
+	}
+    } /* for (round) */
+
+#ifdef HAVE_QSORT
+    /* Sort the matches */
+    qsort((void *)*file, (size_t)count, sizeof(char_u *), compare_mkeys);
+    /* Remove multiple entries */
+    {
+	char_u	**ptr1 = *file;
+	char_u	**ptr2 = ptr1 + 1;
+	char_u	**ptr3 = ptr1 + count;
+
+	while (ptr2 < ptr3)
+	{
+	    if (STRCMP(*ptr1, *ptr2))
+		*++ptr1 = *ptr2++;
+	    else
+	    {
+		vim_free(*ptr2++);
+		count--;
+	    }
+	}
+    }
+#endif /* HAVE_QSORT */
+
+    *num_file = count;
+    return (count == 0 ? FAIL : OK);
+}
+#endif /* CMDLINE_COMPL */
+
 /*
  * Check for an abbreviation.
  * Cursor is at ptr[col]. When inserting, mincol is where insert started.
@@ -2446,8 +2704,11 @@ makemap(fd)
     struct mapblock *mp;
     char_u	    c1, c2;
     char_u	    *p;
+    char	    *cmd;
     int		    abbr;
     int		    hash;
+    int		    did_cpo = FALSE;
+    int		    i;
 
     validate_maphash();
 
@@ -2471,9 +2732,9 @@ makemap(fd)
 		c1 = NUL;
 		c2 = NUL;
 		if (abbr)
-		    p = (char_u *)"abbr";
+		    cmd = "abbr";
 		else
-		    p = (char_u *)"map";
+		    cmd = "map";
 		switch (mp->m_mode)
 		{
 		    case NORMAL + VISUAL + OP_PENDING:
@@ -2501,7 +2762,7 @@ makemap(fd)
 			break;
 		    case CMDLINE + INSERT:
 			if (!abbr)
-			    p = (char_u *)"map!";
+			    cmd = "map!";
 			break;
 		    case CMDLINE:
 			c1 = 'c';
@@ -2515,11 +2776,34 @@ makemap(fd)
 		}
 		do	/* may do this twice if c2 is set */
 		{
+		    /* When outputting <> form, need to make sure that 'cpo'
+		     * is empty. */
+		    if (!did_cpo)
+		    {
+			if (*mp->m_str == NUL)		/* will use <Nop> */
+			    did_cpo = TRUE;
+			else
+			    for (i = 0; i < 2; ++i)
+				for (p = (i ? mp->m_str : mp->m_keys); *p; ++p)
+				    if (*p == K_SPECIAL || *p == NL)
+					did_cpo = TRUE;
+			if (did_cpo)
+			{
+			    if (fprintf(fd,
+#ifdef USE_CRNL
+					"let cpo_save=&cpo\r\nset cpo=\r\n"
+#else
+					"let cpo_save=&cpo\nset cpo=\n"
+#endif
+				       ) < 0)
+				return FAIL;
+			}
+		    }
 		    if (c1 && putc(c1, fd) < 0)
 			return FAIL;
 		    if (mp->m_noremap && fprintf(fd, "nore") < 0)
 			return FAIL;
-		    if (fprintf(fd, (char *)p) < 0)
+		    if (fprintf(fd, cmd) < 0)
 			return FAIL;
 
 		    if (       putc(' ', fd) < 0
@@ -2537,6 +2821,16 @@ makemap(fd)
 		while (c1);
 	    }
 	}
+
+    if (did_cpo)
+	if (fprintf(fd,
+#ifdef USE_CRNL
+		    "let &cpo=cpo_save\r\nunlet cpo_save\r\n"
+#else
+		    "let &cpo=cpo_save\nunlet cpo_save\n"
+#endif
+		    ) < 0)
+	    return FAIL;
     return OK;
 }
 
@@ -2554,6 +2848,13 @@ putescstr(fd, str, set)
     int	    c;
     int	    modifiers;
 
+    /* :map xx <Nop> */
+    if (*str == NUL && !set)
+    {
+	fprintf(fd, "<Nop>");
+	return OK;
+    }
+
     for ( ; *str; ++str)
     {
 	c = *str;
@@ -2598,13 +2899,15 @@ putescstr(fd, str, set)
 	 * prevent them from misinterpreted in DoOneCmd().
 	 * A space, Tab and '"' has to be escaped with a backslash to
 	 * prevent it to be misinterpreted in do_set().
+	 * A '<' has to be escaped with a CTRL-V to prevent it being
+	 * interpreted as the start of a special key name.
 	 */
 	if (set && (vim_iswhite(c) || c == '"' || c == '\\'))
 	{
 	    if (putc('\\', fd) < 0)
 		return FAIL;
 	}
-	else if (c < ' ' || c > '~' || c == '|')
+	else if (c < ' ' || c > '~' || c == '|' || (!set && c == '<'))
 	{
 	    if (putc(Ctrl('V'), fd) < 0)
 		return FAIL;
