@@ -26,7 +26,7 @@
  * Number of Rows and Columns in the screen.
  * Must be long to be able to use them as options in option.c.
  */
-EXTERN long	Rows INIT(= MIN_ROWS + 1);   /* nr of rows in the screen */
+EXTERN long	Rows INIT(= MIN_LINES);		/* nr of rows in the screen */
 EXTERN long	Columns INIT(= MIN_COLUMNS); /* nr of columns in the screen */
 
 /*
@@ -62,6 +62,9 @@ EXTERN int	cmdline_row;
 
 EXTERN int	redraw_cmdline INIT(= FALSE);	/* cmdline must be redrawn */
 EXTERN int	clear_cmdline INIT(= FALSE);	/* cmdline must be cleared */
+#ifdef CRYPTV
+EXTERN int	cmdline_crypt INIT(= FALSE);	/* cmdline is crypted */
+#endif
 EXTERN int	modified INIT(= FALSE);		/* buffer was modified since
 						    last redraw */
 EXTERN int	tag_modified INIT(= FALSE);	/* buffer was modified since
@@ -134,6 +137,9 @@ EXTERN int	intr_char INIT(= 0);	    /* extra interrupt character */
 EXTERN int	vgetc_busy INIT(= FALSE);   /* inside vgetc() now */
 EXTERN int	call_shell_retval;	    /* return value from call_shell() */
 
+EXTERN int	didset_vim INIT(= FALSE);   /* did set $VIM ourselves */
+EXTERN int	didset_vimruntime INIT(= FALSE);   /* idem for $VIMRUNTIME */
+
 /*
  * Lines left before a "more" message.	Ex mode needs to be able to reset this
  * after you type something.
@@ -161,6 +167,7 @@ EXTERN int	autocmd_busy INIT(= FALSE);	/* Is apply_autocmds() busy? */
 EXTERN int	autocmd_no_enter INIT(= FALSE); /* *Enter autocmds disabled */
 EXTERN int	autocmd_no_leave INIT(= FALSE); /* *Leave autocmds disabled */
 EXTERN int	modified_was_set;		/* did ":set modified" */
+EXTERN int	did_filetype INIT(= FALSE);	/* FileType event found */
 #endif
 
 #ifdef USE_MOUSE
@@ -179,6 +186,16 @@ EXTERN int	mouse_past_eol INIT(= FALSE);	/* mouse right of line */
 EXTERN int	WantQueryMouse INIT(= 0);
 #endif
 
+#ifdef USE_GUI
+/* When the window layout is about to be changed, need_mouse_correct is set,
+ * so that gui_mouse_correct() is called afterwards, to correct the mouse
+ * pointer when focus-follow-mouse is being used. */
+EXTERN int	need_mouse_correct INIT(= FALSE);
+
+/* When double clicking, topline must be the same */
+EXTERN linenr_t gui_prev_topline INIT(= 0);
+#endif
+
 #endif
 
 #ifdef USE_GUI
@@ -254,7 +271,12 @@ EXTERN int	silent_mode INIT(= FALSE);
 
 EXTERN FPOS	VIsual;		/* start position of active Visual selection */
 EXTERN int	VIsual_active INIT(= FALSE);
-				/* wheter Visual mode is active */
+				/* whether Visual mode is active */
+EXTERN int	VIsual_select INIT(= FALSE);
+				/* whether Select mode is active */
+EXTERN int	VIsual_reselect;
+				/* whether to restart the selection after a
+				 * Select mode mapping or menu */
 
 EXTERN int	VIsual_mode INIT(= 'v');
 				/* type of Visual mode */
@@ -277,6 +299,16 @@ EXTERN FPOS	where_paste_started;
  */
 EXTERN int     did_ai INIT(= FALSE);
 
+#ifdef SCROLLBIND
+/*
+ * This flag is set after a ":syncbind" to let the check_scrollbind() function
+ * know that it should not attempt to perform scrollbinding due to the scroll
+ * that was a result of the ":syncbind." (Otherwise, check_scrollbind() will
+ * undo some of the work done by ":syncbind.")  -ralston
+ */
+EXTERN int     did_syncbind INIT(= FALSE);
+#endif
+
 #ifdef SMARTINDENT
 /*
  * This flag is set when a smart indent has been performed. When the next typed
@@ -297,6 +329,27 @@ EXTERN int	can_si INIT(= FALSE);
 EXTERN int	can_si_back INIT(= FALSE);
 #endif
 
+#ifdef MULTI_BYTE
+/*
+ * These flags are set based upon 'fileencoding'
+ */
+# define DBCS_JPN    932
+# define DBCS_KOR    949
+# define DBCS_CHS    936
+# define DBCS_CHT    950
+EXTERN int	is_dbcs INIT(= FALSE);	/* One of DBCS_xxx values if DBCS
+					   encoding */
+EXTERN int	is_unicode INIT(= FALSE);
+EXTERN int	is_funky_dbcs INIT(= FALSE);	/* if DBCS encoding, but not CP of system */
+#endif
+
+#ifdef USE_XIM
+EXTERN XIC	xic;
+EXTERN XIM	xim;
+EXTERN GuiColor	xim_fg_color INIT(= -1);
+EXTERN GuiColor	xim_bg_color INIT(= -1);
+#endif
+
 EXTERN int	State INIT(= NORMAL);	/* This is the current state of the
 					 * command interpreter. */
 /*
@@ -340,15 +393,16 @@ EXTERN char_u	*exe_name;		/* the name of the executable */
 #endif
 
 #ifdef USE_GUI_WIN32
-EXTERN	int	dont_scroll INIT(= FALSE);/* don't use scrollbars when TRUE */
+EXTERN int	dont_scroll INIT(= FALSE);/* don't use scrollbars when TRUE */
+EXTERN int	mapped_ctrl_c INIT(= FALSE); /* CTRL-C is mapped */
 #endif
 
 #ifdef USE_BROWSE
-EXTERN int	do_browse INIT(= FALSE);/* TRUE to invoke file dialogue */
+EXTERN int	browse INIT(= FALSE);/* TRUE to invoke file dialog */
 #endif
 
-#if defined(GUI_DIALOGUE) || defined(CON_DIALOGUE)
-EXTERN int	do_confirm INIT(= FALSE);/* TRUE to invoke yes/no dialogue */
+#if defined(GUI_DIALOG) || defined(CON_DIALOG)
+EXTERN int	confirm INIT(= FALSE);/* TRUE to invoke yes/no dialog */
 #endif
 
 EXTERN char_u	*IObuff;		/* sprintf's are done in this buffer */
@@ -408,6 +462,8 @@ EXTERN char_u	*last_cmdline INIT(= NULL); /* last command line (for ":) */
 EXTERN char_u	*new_last_cmdline INIT(= NULL);	/* new value for last_cmdline */
 #ifdef AUTOCMD
 EXTERN char_u	*autocmd_fname INIT(= NULL); /* fname for <afile> on cmdline */
+EXTERN int	autocmd_bufnr INIT(= 0);    /* fnum for <abuf> on cmdline */
+EXTERN char_u	*autocmd_match INIT(= NULL); /* name for <amatch> on cmdline */
 #endif
 
 EXTERN int	postponed_split INIT(= 0);  /* for CTRL-W CTRL-] command */
@@ -454,19 +510,41 @@ extern char *longVersion;		/* this is in version.c */
  * depend on the Makefile.
  */
 #ifdef HAVE_PATHDEF
-extern char_u *default_vim_dir;		/* this is in pathdef.c */
-extern char_u *all_cflags;		/* this is in pathdef.c */
-extern char_u *all_lflags;		/* this is in pathdef.c */
+/* these are in pathdef.c */
+extern char_u *default_vim_dir;
+extern char_u *all_cflags;
+extern char_u *all_lflags;
+extern char_u *compiled_user;
+extern char_u *compiled_sys;
 #endif
 
+/* Characters from 'listchars' option */
+EXTERN int	lcs_eol INIT(= '$');
+EXTERN int	lcs_ext INIT(= NUL);
+EXTERN int	lcs_tab1 INIT(= NUL);
+EXTERN int	lcs_tab2 INIT(= NUL);
+EXTERN int	lcs_trail INIT(= NUL);
+
 EXTERN char_u no_lines_msg[]	INIT(="--No lines in buffer--");
 
+/* table to store parsed 'wildmode' */
+EXTERN char_u	wim_flags[4];
+
+#ifdef EXTRA_SEARCH
+/* don't use 'hlsearch' temporarily */
+EXTERN int	no_hlsearch INIT(= FALSE);
+#endif
+
+#ifdef CURSOR_SHAPE
+/* the table is in misc2.c, because of initializations */
+extern struct cursor_entry cursor_table[SHAPE_COUNT];
+#endif
+
 /*
  * The error messages that can be shared are included here.
  * Excluded are errors that are only used once and debugging messages.
  */
 EXTERN char_u e_abort[]		INIT(="Command aborted");
-EXTERN char_u e_ambmap[]	INIT(="Ambiguous mapping");
 EXTERN char_u e_argreq[]	INIT(="Argument required");
 EXTERN char_u e_backslash[]	INIT(="\\ should be followed by /, ? or &");
 EXTERN char_u e_curdir[]	INIT(="Command not allowed from exrc/vimrc in current dir or tag search");
@@ -485,6 +563,7 @@ EXTERN char_u e_markinval[]	INIT(="Mark has invalid line number");
 EXTERN char_u e_marknotset[]	INIT(="Mark not set");
 EXTERN char_u e_nesting[]	INIT(="Scripts nested too deep");
 EXTERN char_u e_noalt[]		INIT(="No alternate file");
+EXTERN char_u e_noabbr[]	INIT(="No such abbreviation");
 EXTERN char_u e_nobang[]	INIT(="No ! allowed");
 EXTERN char_u e_nogvim[]	INIT(="GUI cannot be used: Not enabled at compile time\n");
 EXTERN char_u e_nohebrew[]	INIT(="Hebrew cannot be used: Not enabled at compile time\n");
