@@ -32,7 +32,9 @@ static Widget menuBar;
 
 static void scroll_cb __ARGS((Widget w, XtPointer client_data, XtPointer call_data));
 static void gui_motif_add_actext __ARGS((GuiMenu *menu));
+#if (XmVersion >= 1002)
 static void gui_mch_recurse_tearoffs __ARGS((GuiMenu *menu, int val));
+#endif
 static void gui_mch_compute_menu_height __ARGS((Widget));
 static void gui_mch_submenu_colors __ARGS((GuiMenu *mp));
 
@@ -61,7 +63,14 @@ scroll_cb(w, client_data, call_data)
     value = ((XmScrollBarCallbackStruct *)call_data)->value;
     dragging = (((XmScrollBarCallbackStruct *)call_data)->reason ==
 							      (int)XmCR_DRAG);
-    gui_drag_scrollbar(sb, value, dragging);
+    /*
+     * :NOTE: 1998-11-18 09:26:01 EST eralston
+     * pretend we're not dragging whenever scrollbind is active in
+     * the current window, so that the other scrollbind windows
+     * get their scrollbars updated also.
+     */
+    gui_drag_scrollbar(sb, value, dragging
+		    && (!sb->wp || !sb->wp->w_p_scbind || (sb->wp != curwin)));
 }
 
 
@@ -349,14 +358,17 @@ gui_mch_toggle_tearoffs(enable)
 {
     int		val;
 
+#if (XmVersion >= 1002)
     if (enable)
 	val = (int)XmTEAR_OFF_ENABLED;
     else
 	val = (int)XmTEAR_OFF_DISABLED;
     XtVaSetValues(menuBar, XmNtearOffModel, val, NULL);
     gui_mch_recurse_tearoffs(gui.root_menu, val);
+#endif
 }
 
+#if (XmVersion >= 1002)
     static void
 gui_mch_recurse_tearoffs(menu, val)
     GuiMenu	*menu;
@@ -373,6 +385,7 @@ gui_mch_recurse_tearoffs(menu, val)
 	menu = menu->next;
     }
 }
+#endif
 
 /*
  * Compute the height of the menu bar.
@@ -461,6 +474,10 @@ gui_mch_add_menu_item(menu, parent, idx)
     if (parent->submenu_id == (Widget)0)
 	return;
 
+    /* Don't add menu separator */
+    if (gui_is_separator(menu->name))
+	return;
+
     label = XmStringCreate((char *)menu->dname, STRING_TAG);
     if (label == NULL)
 	return;
