@@ -15,11 +15,24 @@
 
 #include "vim.h"
 
+#ifdef USE_FONTSET
+# ifdef X_LOCALE
+#  include <X11/Xlocale.h>
+# else
+#  include <locale.h>
+# endif
+#endif
+
 #define VIM_NAME	"vim"
 #define VIM_CLASS	"Vim"
 
 /* Default resource values */
-#define DFLT_FONT		"7x13"
+#ifdef USE_FONTSET
+/* It works for me. but any others? */
+# define DFLT_FONT		"7x14,-*-*-*-*-*--*-*-*-*-c-140-*-*"
+#else
+# define DFLT_FONT		"7x13"
+#endif
 #define DFLT_MENU_BG_COLOR	"gray77"
 #define DFLT_MENU_FG_COLOR	"black"
 #define DFLT_SCROLL_BG_COLOR	"gray60"
@@ -33,6 +46,15 @@ static Atom   wm_atoms[2];	/* Window Manager Atoms */
 #define DELETE_WINDOW_IDX 0	/* index in wm_atoms[] for WM_DELETE_WINDOW */
 #define SAVE_YOURSELF_IDX 1	/* index in wm_atoms[] for WM_SAVE_YOURSELF */
 
+#ifdef USE_FONTSET
+static XFontSet current_fontset;
+#define XDrawString(dpy,win,gc,x,y,str,n) \
+	XmbDrawString(dpy,win,current_fontset,gc,x,y,str,n)
+static GuiColor	prev_fg_color = (GuiColor) -1;
+static GuiColor	prev_bg_color = (GuiColor) -1;
+static Window workwid = 0;
+#endif
+
 static void gui_x11_timer_cb __ARGS((XtPointer timed_out, XtIntervalId *interval_id));
 static void gui_x11_visibility_cb __ARGS((Widget w, XtPointer dud, XEvent *event, Boolean *dum));
 static void gui_x11_expose_cb __ARGS((Widget w, XtPointer dud, XEvent *event, Boolean *dum));
@@ -51,7 +73,6 @@ static void  clip_x11_lose_ownership_cb __ARGS((Widget, Atom *));
 static void  gui_x11_wm_protocol_handler __ARGS((Widget, XtPointer, XEvent *, Boolean *));
 static void gui_x11_blink_cb __ARGS((XtPointer timed_out, XtIntervalId *interval_id));
 static Cursor gui_x11_create_blank_mouse __ARGS((void));
-static void gui_x11_blank_mouse __ARGS((int should_blank));
 
 
 static struct specialkey
@@ -297,6 +318,35 @@ static XtResource vim_resources[] =
 	XtRString,
 	DFLT_SCROLL_BG_COLOR
     },
+#ifdef USE_XIM
+    {
+	"preeditType",
+	"PreeditType",
+	XtRString,
+	sizeof(char*),
+	XtOffsetOf(Gui, preedit_type),
+	XtRString,
+	(XtPointer)"OverTheSpot"
+    },
+    {
+	"inputMethod",
+	"InputMethod",
+	XtRString,
+	sizeof(char*),
+	XtOffsetOf(Gui, input_method),
+	XtRString,
+	NULL
+    },
+    {
+	"openIM",
+	"OpenIM",
+	XtRBoolean,
+	sizeof(Boolean),
+	XtOffsetOf(Gui, open_im),
+	XtRImmediate,
+	(XtPointer)TRUE
+    },
+#endif /* USE_XIM */
 };
 
 /*
@@ -337,6 +387,110 @@ static XrmOptionDescRec cmdline_options[] =
 static int gui_argc = 0;
 static char **gui_argv = NULL;
 
+#ifdef USE_FONTSET
+    static void
+create_workwindow()
+{
+    Window	    root;
+    unsigned int    dummy;
+    unsigned int    depth;
+
+    if (workwid)
+	XFreePixmap (gui.dpy, workwid);
+    XGetGeometry(gui.dpy, gui.wid, &root, &dummy, &dummy, &dummy, &dummy,
+		 &dummy, &depth);
+    workwid = XCreatePixmap(gui.dpy, gui.wid,
+			    Columns * gui.char_width, gui.char_height,
+			    depth);
+}
+
+    static int
+check_fontset_sanity(fs)
+    XFontSet fs;
+{
+    XFontStruct **xfs;
+    char **font_name;
+    int fn;
+    char *base_name;
+
+    base_name = XBaseFontNameListOfFontSet(fs);
+    fn = XFontsOfFontSet(fs, &xfs, &font_name);
+    if (fn != 2)
+    {
+	char *locale = XLocaleOfFontSet(fs);
+
+	if (!locale || locale[0] == 'C')
+	{
+	    EMSG("locale is not set correctly\n"
+		 "Set LANG environment variable to your locale\n"
+		 "For korean:\n"
+		 "   csh: setenv LANG ko\n"
+		 "   sh : export LANG=ko\n");
+	}
+	EMSG2("fontset name: %s\n", base_name);
+	EMSG("Your language Font missing\n");
+	EMSG2("loaded fontname: %s\n\n", font_name[0]);
+	return 1;
+    }
+    if (xfs[0]->max_bounds.width != xfs[0]->min_bounds.width) 
+    {
+	EMSG2("Fontset name: %s\n", base_name);
+	EMSG2("Font '%s' is not fixed-width\n\n", font_name[0]);
+	return 1;
+    }
+    if (xfs[1]->max_bounds.width != 2*xfs[0]->max_bounds.width)
+    {
+	EMSG2("Fontset name: %s\n", base_name);
+	EMSG2("Font0: %s\n", font_name[0]);
+	EMSG2("Font1: %s\n", font_name[1]);
+	EMSG("Font1 width is not twice that of font0\n");
+	EMSGN("Font0 width: %ld\n", xfs[0]->max_bounds.width);
+	EMSGN("Font1 width: %ld\n\n", xfs[1]->max_bounds.width);
+	return 1;
+    }
+    /* it seems ok. Good Luck!! */
+    return 0;
+}
+
+    static int
+fontset_width(XFontSet fs)
+{
+    return XmbTextEscapement(fs, "Vim", 3) / 3;
+}
+
+    static int
+fontset_height(fs)
+    XFontSet fs;
+{
+    XFontSetExtents *extents;
+
+    extents = XExtentsOfFontSet(fs);
+    return extents->max_logical_extent.height;
+}
+
+/* NOT USED YET
+    static int
+fontset_descent(fs)
+    XFontSet fs;
+{
+    XFontSetExtents *extents;
+
+    extents = XExtentsOfFontSet (fs);
+    return extents->max_logical_extent.height + extents->max_logical_extent.y;
+}
+*/
+
+    static int
+fontset_ascent(fs)
+    XFontSet fs;
+{
+    XFontSetExtents *extents;
+
+    extents = XExtentsOfFontSet (fs);
+    return -extents->max_logical_extent.y;
+}
+#endif
+
 /*
  * Call-back routines.
  */
@@ -411,7 +565,16 @@ gui_x11_resize_window_cb(w, dud, event, dum)
     if (event->type != ConfigureNotify)
 	return;
 
+#ifndef USE_XIM
     gui_resize_window(event->xconfigure.width, event->xconfigure.height);
+#else
+    gui_resize_window(event->xconfigure.width, event->xconfigure.height
+		                               - gui.char_height);
+    xim_set_preedit();
+#endif
+#ifdef USE_FONTSET
+    create_workwindow ();
+#endif
 }
 
 /* ARGSUSED */
@@ -423,6 +586,9 @@ gui_x11_focus_change_cb(w, data, event, dum)
     Boolean	*dum;
 {
     gui_focus_change(event->type == FocusIn);
+#ifdef USE_XIM
+    xim_set_focus(event->type == FocusIn);
+#endif
 }
 
 /* ARGSUSED */
@@ -434,6 +600,7 @@ gui_x11_enter_cb(w, data, event, dum)
     Boolean	*dum;
 {
     gui_focus_change(TRUE);
+    gui_update_cursor(TRUE, FALSE);
 }
 
 /* ARGSUSED */
@@ -445,6 +612,7 @@ gui_x11_leave_cb(w, data, event, dum)
     Boolean	*dum;
 {
     gui_focus_change(FALSE);
+    gui_update_cursor(TRUE, FALSE);
 }
 
 /* ARGSUSED */
@@ -456,7 +624,12 @@ gui_x11_key_hit_cb(w, dud, event, dum)
     Boolean	*dum;
 {
     XKeyPressedEvent	*ev_press;
+#ifdef USE_XIM
+    char_u		string[256], string2[256];
+    Status		status;
+#else
     char_u		string[3], string2[3];
+#endif
     KeySym		key_sym, key_sym2;
     int			len, len2;
     int			i;
@@ -465,8 +638,14 @@ gui_x11_key_hit_cb(w, dud, event, dum)
 
     ev_press = (XKeyPressedEvent *)event;
 
-    len = XLookupString(ev_press, (char *)string, sizeof(string),
-	    &key_sym, NULL);
+#ifdef USE_XIM
+    if (xic)
+	len = XmbLookupString(xic, ev_press, string, sizeof(string), &key_sym,
+			      &status);
+    else
+#endif
+	len = XLookupString(ev_press, (char *)string, sizeof(string),
+		&key_sym, NULL);
 
     if (key_sym == XK_space)
 	string[0] = ' ';	/* Otherwise Ctrl-Space doesn't work */
@@ -485,6 +664,12 @@ gui_x11_key_hit_cb(w, dud, event, dum)
     if (len == 1 && (ev_press->state & Mod1Mask)
 			&& !(key_sym == XK_BackSpace || key_sym == XK_Delete))
     {
+#ifdef USE_GUI_MOTIF
+	/* Ignore ALT keys when they are used for the menu only */
+	if (p_wak[0] == 'y'
+		|| (p_wak[0] == 'm' && gui_is_menu_shortcut(string[0])))
+	    return;
+#endif
 	/*
 	 * Before we set the 8th bit, check to make sure the user doesn't
 	 * already have a mapping defined for this sequence. We determine this
@@ -546,26 +731,37 @@ gui_x11_key_hit_cb(w, dud, event, dum)
 	if (ev_press->state & Mod1Mask)
 	    modifiers |= MOD_MASK_ALT;
 
-	/*
-	 * For some keys a shift modifier is translated into another key code.
-	 * Do we need to handle the case where len != 1 and string[0] != CSI?
+#if defined(USE_XIM) && defined(MULTI_BYTE)
+	/* FIXME: I doubt about it. I don't know fully original code.
+	 * But, CSI can be multibyte char. We should avoid if it is a
+	 * real CSI or multi-byte char. This is a quick hack and not
+	 * fully tested.
 	 */
-	if (string[0] == CSI && len == 3)
-	    key = TO_SPECIAL(string[1], string[2]);
-	else
-	    key = string[0];
-	key = simplify_key(key, &modifiers);
-	if (IS_SPECIAL(key))
+	if (!is_dbcs || key_sym)
+#endif
 	{
-	    string[0] = CSI;
-	    string[1] = K_SECOND(key);
-	    string[2] = K_THIRD(key);
-	    len = 3;
-	}
-	else
-	{
-	    string[0] = key;
-	    len = 1;
+	    /*
+	     * For some keys a shift modifier is translated into another key
+	     * code.  Do we need to handle the case where len != 1 and
+	     * string[0] != CSI?
+	     */
+	    if (string[0] == CSI && len == 3)
+		key = TO_SPECIAL(string[1], string[2]);
+	    else
+		key = string[0];
+	    key = simplify_key(key, &modifiers);
+	    if (IS_SPECIAL(key))
+	    {
+		string[0] = CSI;
+		string[1] = K_SECOND(key);
+		string[2] = K_THIRD(key);
+		len = 3;
+	    }
+	    else
+	    {
+		string[0] = key;
+		len = 1;
+	    }
 	}
 
 	if (modifiers)
@@ -590,11 +786,8 @@ gui_x11_key_hit_cb(w, dud, event, dum)
     /*
      * blank out the pointer if necessary
      */
-    if (p_mh && !gui.pointer_hidden)
-    {
-	gui_x11_blank_mouse(TRUE);
-	gui.pointer_hidden = TRUE;
-    }
+    if (p_mh)
+	gui_mch_mousehide(TRUE);
 }
 
 /* ARGSUSED */
@@ -625,14 +818,12 @@ gui_x11_mouse_cb(w, dud, event, dum)
 	/*
 	 * if our pointer is currently hidden, then we should show it.
 	 */
-	if (gui.pointer_hidden)
-	{
-	    gui_x11_blank_mouse(FALSE);
-	    gui.pointer_hidden = FALSE;
-	}
+	gui_mch_mousehide(FALSE);
 
 	if (button != MOUSE_DRAG)	/* just moving the rodent */
 	{
+	    if (dud)			/* moved in vimForm */
+		y -= gui.menu_height;
 	    gui_mouse_moved(y);
 	    return;
 	}
@@ -742,14 +933,14 @@ gui_mch_prepare(argc, argv)
 	    gui_argv[gui_argc++] = argv[arg];
 	    if (--*argc > arg)
 	    {
-		vim_memmove(&argv[arg], &argv[arg + 1], (*argc - arg)
+		mch_memmove(&argv[arg], &argv[arg + 1], (*argc - arg)
 						    * sizeof(char *));
 		if (cmdline_options[i].argKind != XrmoptionNoArg)
 		{
 		    /* Move the options argument as well */
 		    gui_argv[gui_argc++] = argv[arg];
 		    if (--*argc > arg)
-			vim_memmove(&argv[arg], &argv[arg + 1], (*argc - arg)
+			mch_memmove(&argv[arg], &argv[arg + 1], (*argc - arg)
 							    * sizeof(char *));
 		}
 	    }
@@ -772,6 +963,12 @@ gui_mch_init()
     int		x, y, mask;
     unsigned	w, h;
 
+#ifdef USE_FONTSET
+    setlocale(LC_ALL, "");
+#endif
+#ifdef USE_XIM
+    XtSetLanguageProc(NULL, NULL, NULL);
+#endif
     XtToolkitInitialize();
     app_context = XtCreateApplicationContext();
     gui.dpy = XtOpenDisplay(app_context, 0, VIM_NAME, VIM_CLASS,
@@ -904,6 +1101,32 @@ gui_mch_init()
     }
 
     gui_x11_create_widgets();
+
+   /*
+    * Add an icon to Vim (Marcel Douben: 11 May 1998).
+    */
+    if (vim_strchr(p_guioptions, GO_ICON) != NULL)
+    {
+#include "vim_icon.xbm"
+#include "vim_mask.xbm"
+
+	Arg	arg[2];
+
+	XtSetArg(arg[0], XtNiconPixmap,
+		XCreateBitmapFromData(gui.dpy,
+		    DefaultRootWindow(gui.dpy),
+		    (char *)vim_icon_bits,
+		    vim_icon_width,
+		    vim_icon_height));
+	XtSetArg(arg[1], XtNiconMask,
+		XCreateBitmapFromData(gui.dpy,
+		    DefaultRootWindow(gui.dpy),
+		    (char *)vim_mask_icon_bits,
+		    vim_mask_icon_width,
+		    vim_mask_icon_height));
+	XtSetValues(vimShell, arg, (Cardinal)2);
+    }
+
     return OK;
 }
 
@@ -971,11 +1194,17 @@ gui_mch_open()
 		(XtPointer)NULL);
     }
 #endif
+#ifdef USE_XIM
+    xim_init();
+#endif
 
 #ifdef USE_GUI_ATHENA
     /* The Athena GUI needs this again after opening the window */
     gui_position_menu();
 #endif
+#ifdef USE_FONTSET
+    create_workwindow();
+#endif
 
     return OK;
 }
@@ -985,7 +1214,9 @@ gui_mch_open()
 gui_mch_exit(rc)
     int		rc;
 {
-#ifndef LESSTIF_VERSION	    /* Lesstif gives an error message here */
+#if 0
+    /* Lesstif gives an error message here, and so does Solaris.  The man page
+     * says that this isn't needed when exiting, so just skip it. */
     XtCloseDisplay(gui.dpy);
 #endif
 }
@@ -1010,7 +1241,12 @@ gui_mch_set_winsize(width, height, min_width, min_height,
 	XtNminWidth,	min_width,
 	XtNminHeight,	min_height,
 	XtNwidth,	width,
+#if defined(USE_XIM)
+	XtNheight,	height + gui.char_height,
+			/* FIXME (xic?gui.char_height:0) */
+#else
 	XtNheight,	height,
+#endif
 	NULL);
 }
 
@@ -1035,7 +1271,11 @@ gui_mch_get_screen_dimensions(screen_w, screen_h)
 gui_mch_init_font(font_name)
     char_u	*font_name;
 {
+#ifdef USE_FONTSET
+    XFontSet	 font = NULL;
+#else
     XFontStruct	*font = NULL;
+#endif
 
     if (font_name == NULL)
     {
@@ -1045,21 +1285,37 @@ gui_mch_init_font(font_name)
 	 * hopefully always be there.
 	 */
 	font_name = gui.dflt_font;
+#ifdef USE_FONTSET
+	font = (XFontSet)gui_mch_get_font(font_name, FALSE);
+#else
 	font = (XFontStruct *)gui_mch_get_font(font_name, FALSE);
+#endif
 	if (font == NULL)
 	    font_name = (char_u *)DFLT_FONT;
     }
     if (font == NULL)
+#ifdef USE_FONTSET
+	font = (XFontSet)gui_mch_get_font(font_name, FALSE);
+#else
 	font = (XFontStruct *)gui_mch_get_font(font_name, FALSE);
+#endif
     if (font == NULL)
 	return FAIL;
 
     if (gui.norm_font != 0)
+#ifdef USE_FONTSET
+	XFreeFontSet(gui.dpy, (XFontSet)gui.norm_font);
+    gui.norm_font = (GuiFont)font;
+    gui.char_width = fontset_width(font);
+    gui.char_height = fontset_height(font);
+    gui.char_ascent = fontset_ascent(font);
+#else
 	XFreeFont(gui.dpy, (XFontStruct *)gui.norm_font);
     gui.norm_font = (GuiFont)font;
     gui.char_width = font->max_bounds.width;
     gui.char_height = font->ascent + font->descent;
     gui.char_ascent = font->ascent;
+#endif
     hl_set_font_name(font_name);
 
     /*
@@ -1091,12 +1347,50 @@ gui_mch_get_font(name, giveErrorIfMissing)
     char_u	*name;
     int		giveErrorIfMissing;
 {
+#ifdef USE_FONTSET
+    XFontSet	 font;
+#else
     XFontStruct	*font;
+#endif
 
     if (!gui.in_use)		    /* can't do this when GUI not running */
 	return (GuiFont)0;
 
+#ifdef USE_FONTSET
+    {
+	char	**missing, *def_str, *new_name;
+	int	num_missing;
+	int	ti;
+
+	/* User can set fontset by
+	 *    :set guifont=english_font;his_language_font
+	 */
+	new_name = (char *) alloc(strlen(name) + 1);
+	for (ti = 0; name[ti]; ti++)
+	    new_name[ti] = (name[ti] == ';') ? ',' : name[ti];
+	new_name[ti] = '\0';
+
+	font = XCreateFontSet(gui.dpy, new_name, &missing, &num_missing,
+			      &def_str);
+	if (num_missing > 0)
+	{
+	    int i;
+
+	    if (giveErrorIfMissing)
+	    {
+		EMSG2("Error: During loading fontset %s\n", new_name);
+		EMSG("Font(s) for the following character "
+		     "sets are missing:\n");
+		for (i = 0; i < num_missing; i++)
+		    EMSG2("%s\n", missing[i]);
+	    }
+	    XFreeStringList(missing);
+	}
+	vim_free(new_name);
+    }
+#else
     font = XLoadQueryFont(gui.dpy, (char *)name);
+#endif
 
     if (font == NULL)
     {
@@ -1123,12 +1417,20 @@ gui_mch_get_font(name, giveErrorIfMissing)
     printf("\n");
 #endif
 
+#ifdef USE_FONTSET
+    if (check_fontset_sanity(font))
+    {
+	XFreeFontSet(gui.dpy, font);
+	return (GuiFont)0;
+    }
+#else
     if (font->max_bounds.width != font->min_bounds.width)
     {
 	EMSG2("Font \"%s\" is not fixed-width", name);
 	XFreeFont(gui.dpy, font);
 	return (GuiFont)0;
     }
+#endif
     return (GuiFont)font;
 }
 
@@ -1139,6 +1441,9 @@ gui_mch_get_font(name, giveErrorIfMissing)
 gui_mch_set_font(font)
     GuiFont	font;
 {
+#ifdef USE_FONTSET
+    current_fontset = (XFontSet)font;
+#else
     static Font prev_font = (Font) -1;
 
     Font    fid = ((XFontStruct *)font)->fid;
@@ -1149,6 +1454,7 @@ gui_mch_set_font(font)
 	XSetFont(gui.dpy, gui.back_gc, fid);
 	prev_font = fid;
     }
+#endif
 }
 
 /*
@@ -1159,7 +1465,11 @@ gui_mch_same_font(f1, f2)
     GuiFont	f1;
     GuiFont	f2;
 {
+#ifdef USE_FONTSET
+    return f1 == f2;
+#else
     return ((XFontStruct *)f1)->fid == ((XFontStruct *)f2)->fid;
+#endif
 }
 
 /*
@@ -1169,7 +1479,12 @@ gui_mch_same_font(f1, f2)
 gui_mch_free_font(font)
     GuiFont	font;
 {
-    XFreeFont(gui.dpy, (XFontStruct *)font);
+    if (font)
+#ifdef USE_FONTSET
+    	XFreeFontSet(gui.dpy, (XFontSet)font);
+#else
+	XFreeFont(gui.dpy, (XFontStruct *)font);
+#endif
 }
 
 /*
@@ -1232,6 +1547,35 @@ gui_mch_get_color(name)
     return (GuiColor)-1;
 }
 
+#ifdef USE_FONTSET
+
+    void
+gui_mch_set_fg_color(color)
+    GuiColor	color;
+{
+    if (color != prev_fg_color)
+    {
+	XSetForeground(gui.dpy, gui.text_gc, (Pixel)color);
+	prev_fg_color = color;
+    }
+}
+
+/*
+ * Set the current text background color.
+ */
+    void
+gui_mch_set_bg_color(color)
+    GuiColor	color;
+{
+    if (color != prev_bg_color)
+    {
+	XSetBackground(gui.dpy, gui.text_gc, (Pixel)color);
+	prev_bg_color = color;
+    }
+}
+
+#else /* !USE_FONTSET */
+
 /*
  * Set the current text foreground color.
  */
@@ -1263,7 +1607,7 @@ gui_mch_set_bg_color(color)
 	prev_color = color;
     }
 }
-
+#endif /* !USE_FONTSET */
 /*
  * create a mouse pointer that is blank
  */
@@ -1279,16 +1623,20 @@ gui_x11_create_blank_mouse()
 }
 
 /*
- * use the blank mouse pointer
+ * Use the blank mouse pointer or not.
  */
-    static void
-gui_x11_blank_mouse(should_blank)
-    int should_blank;  /* nonzero = use blank ptr ; zero = use parent ptr */
+    void
+gui_mch_mousehide(hide)
+    int		hide;	/* TRUE = use blank ptr, FALSE = use parent ptr */
 {
-    if (should_blank)
-	XDefineCursor(gui.dpy, gui.wid, gui.blank_pointer);
-    else
-	XUndefineCursor(gui.dpy, gui.wid);
+    if (gui.pointer_hidden != hide)
+    {
+	if (hide)
+	    XDefineCursor(gui.dpy, gui.wid, gui.blank_pointer);
+	else
+	    XUndefineCursor(gui.dpy, gui.wid);
+	gui.pointer_hidden = hide;
+    }
 }
 
     void
@@ -1299,12 +1647,46 @@ gui_mch_draw_string(row, col, s, len, flags)
     int	    len;
     int	    flags;
 {
+#ifdef USE_FONTSET
+    if (col > 0 && mb_isbyte1(LinePointers[row], col - 1))
+    {
+	col++;
+	len--;
+	s++;
+    }
+    if ((len == 1) && (*s & 0x80))
+	len = 2;
+#endif
     if (flags & DRAW_TRANSP)
 	XDrawString(gui.dpy, gui.wid, gui.text_gc, TEXT_X(col),
 	    TEXT_Y(row), (char *)s, len);
     else
+#ifdef USE_FONTSET
+    {
+#ifdef SLOW_XSERVER
+	/* XmbDrawImageString has bug. Don't use it */
+	XSetForeground(gui.dpy, gui.text_gc, prev_bg_color);
+	XFillRectangle(gui.dpy, gui.wid, gui.text_gc, FILL_X(col),
+		       FILL_Y(row), gui.char_width * len, gui.char_height);
+	XSetForeground(gui.dpy, gui.text_gc, prev_fg_color);
+	XDrawString(gui.dpy, gui.wid, gui.text_gc, TEXT_X(col),
+		    TEXT_Y(row), (char *)s, len);
+#else /* !SLOW_XSERVER */
+	XSetForeground(gui.dpy, gui.text_gc, prev_bg_color);
+	XFillRectangle(gui.dpy, workwid, gui.text_gc, 0, 0,
+		       gui.char_width * len, gui.char_height);
+	XSetForeground (gui.dpy, gui.text_gc, prev_fg_color);
+	XDrawString(gui.dpy, workwid, gui.text_gc, 0, gui.char_ascent,
+		    (char *)s, len);
+	XCopyArea (gui.dpy, workwid, gui.wid, gui.text_gc, 0, 0,
+		   gui.char_width *len, gui.char_height,
+		   FILL_X(col), FILL_Y(row));
+#endif /* !SLOW_XSERVER */
+    }
+#else
 	XDrawImageString(gui.dpy, gui.wid, gui.text_gc, TEXT_X(col),
 	    TEXT_Y(row), (char *)s, len);
+#endif
 
     if (flags & DRAW_BOLD)
 	XDrawString(gui.dpy, gui.wid, gui.text_gc, TEXT_X(col) + 1,
@@ -1394,9 +1776,18 @@ gui_mch_iconify()
 gui_mch_draw_hollow_cursor(color)
     GuiColor color;
 {
+#ifdef USE_FONTSET
+    if (LinePointers[gui.row][gui.col] & 0x80)
+	XDrawRectangle(gui.dpy, gui.wid, gui.text_gc, FILL_X(gui.col),
+		   FILL_Y(gui.row), 2*gui.char_width - 1, gui.char_height - 1);
+    else
+	XDrawRectangle(gui.dpy, gui.wid, gui.text_gc, FILL_X(gui.col),
+		   FILL_Y(gui.row), gui.char_width - 1, gui.char_height - 1);
+#else
     gui_mch_set_fg_color(color);
     XDrawRectangle(gui.dpy, gui.wid, gui.text_gc, FILL_X(gui.col),
 		   FILL_Y(gui.row), gui.char_width - 1, gui.char_height - 1);
+#endif
 }
 
 /*
@@ -1786,11 +2177,11 @@ clip_x11_convert_selection_cb(w, selection, target, type, value, length, format)
 	return False;
     }
     if (*target == XA_STRING)
-	vim_memmove(result, string, (size_t)(*length));
+	mch_memmove(result, string, (size_t)(*length));
     else
     {
 	result[0] = motion_type;
-	vim_memmove(result + 1, string, (size_t)(*length - 1));
+	mch_memmove(result + 1, string, (size_t)(*length - 1));
     }
     *type = *target;
     *format = 8;	    /* 8 bits per char */
@@ -1845,11 +2236,14 @@ gui_mch_menu_grey(menu, grey)
     GuiMenu *menu;
     int	    grey;
 {
-    gui_mch_menu_hidden(menu, False);
-    if (grey)
-	XtSetSensitive(menu->id, False);
-    else
-	XtSetSensitive(menu->id, True);
+    if (menu->id != (Widget)0)
+    {
+	gui_mch_menu_hidden(menu, False);
+	if (grey)
+	    XtSetSensitive(menu->id, False);
+	else
+	    XtSetSensitive(menu->id, True);
+    }
 }
 
 /*
@@ -1860,10 +2254,13 @@ gui_mch_menu_hidden(menu, hidden)
     GuiMenu *menu;
     int	    hidden;
 {
-    if (hidden)
-	XtUnmanageChild(menu->id);
-    else
-	XtManageChild(menu->id);
+    if (menu->id != (Widget)0)
+    {
+	if (hidden)
+	    XtUnmanageChild(menu->id);
+	else
+	    XtManageChild(menu->id);
+    }
 }
 
 /*
@@ -1893,10 +2290,13 @@ gui_mch_enable_scrollbar(sb, flag)
     GuiScrollbar    *sb;
     int		    flag;
 {
-    if (flag)
-	XtManageChild(sb->id);
-    else
-	XtUnmanageChild(sb->id);
+    if (sb->id != (Widget)0)
+    {
+	if (flag)
+	    XtManageChild(sb->id);
+	else
+	    XtUnmanageChild(sb->id);
+    }
 }
 
 
@@ -1936,7 +2336,7 @@ gui_x11_wm_protocol_handler(w, client_data, event, dum)
 
     /* Only exit when there are no changed files */
     exiting = TRUE;
-    if (!check_changed_any())	    /* will give warning for changed buffer */
+    if (!check_changed_any(FALSE))    /* will give warning for changed buffer */
 	getout(0);
 
     exiting = FALSE;
@@ -2110,7 +2510,55 @@ gui_x11_callbacks(textArea, vimForm)
     XtAddEventHandler(textArea, KeyPressMask, FALSE, gui_x11_key_hit_cb,
 	(XtPointer)0);
 
+    /* get pointer moved events from scrollbar, needed for 'mousefocus' */
+    XtAddEventHandler(vimForm, PointerMotionMask,
+	FALSE, gui_x11_mouse_cb, (XtPointer)1);
     XtAddEventHandler(textArea, ButtonPressMask | ButtonReleaseMask |
 					 ButtonMotionMask | PointerMotionMask,
 	FALSE, gui_x11_mouse_cb, (XtPointer)0);
 }
+
+/*
+ * Get current y mouse coordinate in text window.
+ * Return -1 when unknown.
+ */
+    int
+gui_mch_get_mouse_x()
+{
+    int		rootx, rooty, winx, winy;
+    Window	root, child;
+    unsigned int mask;
+
+    if (XQueryPointer(gui.dpy, XtWindow(vimShell), &root, &child,
+		&rootx, &rooty, &winx, &winy, &mask))
+    {
+	if (gui.which_scrollbars[SBAR_LEFT])
+	    return winx - gui.scrollbar_width;
+	return winx;
+    }
+    return -1;
+}
+
+    int
+gui_mch_get_mouse_y()
+{
+    int		rootx, rooty, winx, winy;
+    Window	root, child;
+    unsigned int mask;
+
+    if (XQueryPointer(gui.dpy, XtWindow(vimShell), &root, &child,
+		&rootx, &rooty, &winx, &winy, &mask))
+	return winy - gui.menu_height;
+    return -1;
+}
+
+    void
+gui_mch_setmouse(x, y)
+    int		x;
+    int		y;
+{
+    if (gui.which_scrollbars[SBAR_LEFT])
+	x += gui.scrollbar_width;
+    XWarpPointer(gui.dpy, (Window)0, XtWindow(vimShell), 0, 0, 0, 0,
+						      x, y + gui.menu_height);
+}
