@@ -40,7 +40,8 @@ static int	    cs_find __ARGS((EXARG *eap));
 static int	    cs_find_common __ARGS((char *opt, char *pat, int, int ));
 static int	    cs_help __ARGS((EXARG *eap));
 static void	    cs_init __ARGS((void));
-static int	    cs_insert_filelist __ARGS((char *, char *, char *));
+static int	    cs_insert_filelist __ARGS((char *, char *, char *,
+			struct stat *));
 static int	    cs_kill __ARGS((EXARG *eap));
 static cscmd_t *    cs_lookup_cmd __ARGS((EXARG *eap));
 static char *	    cs_make_vim_style_matches __ARGS((char *, char *,
@@ -284,7 +285,7 @@ cs_add_common(arg1, arg2, flags)
     ret = stat(fname, &statbuf);
     if (ret < 0) {
 	if (p_csverbose) {
-	    (void)sprintf(buf, "stat(%s) error: %d)", fname, errno);
+	    (void)sprintf(buf, "stat(%s) error: %d", fname, errno);
 	    (void)EMSG(buf);
 	}
 	goto add_err;
@@ -313,7 +314,7 @@ cs_add_common(arg1, arg2, flags)
     /* if filename is a directory, append the cscope database name to it */
     if ((statbuf.st_mode & S_IFMT) == S_IFDIR) {
 	char *fname2;
-	fname2 = (char *)alloc(strlen(CSCOPE_DBFILE) + strlen(fname));
+	fname2 = (char *)alloc(strlen(CSCOPE_DBFILE) + strlen(fname) + 2);
 	if (fname2 == NULL) {
 	    (void)EMSG("cs_add_common: alloc fail #4");
 	    goto add_err;
@@ -338,12 +339,12 @@ cs_add_common(arg1, arg2, flags)
 	    vim_free(fname2);
 	    goto add_err;
 	}
-	i = cs_insert_filelist(fname2, ppath, flags);
+	i = cs_insert_filelist(fname2, ppath, flags, &statbuf);
 	if (p_csverbose)
 	    (void)sprintf(buf, "Added cscope database %s", fname2);
 	vim_free(fname2);
-    } else if ((statbuf.st_mode & S_IFMT) == S_IFREG) {
-	i = cs_insert_filelist(fname, ppath, flags);
+    } else if (S_ISREG(statbuf.st_mode) || S_ISLNK(statbuf.st_mode)) {
+	i = cs_insert_filelist(fname, ppath, flags, &statbuf);
 	if (p_csverbose)
 	    (void)sprintf(buf, "Added cscope database %s", fname);
     } else {
@@ -655,7 +656,8 @@ cs_find(eap)
 	return FALSE;
     }
 
-    if ((pat = strtok((char *)NULL, (const char *)" ")) == NULL) {
+    pat = opt + strlen(opt) + 1;
+    if (pat == NULL || (pat != NULL && pat[0] == '\0')) {
 	CS_USAGE_MSG(Find);
 	return FALSE;
     }
@@ -786,12 +788,14 @@ cs_init()
 	return;
 
     for (i = 0; i < CSCOPE_MAX_CONNECTIONS; i++) {
-	csinfo[i].fname = NULL;
-	csinfo[i].ppath = NULL;
-	csinfo[i].flags = NULL;
-	csinfo[i].pid   = -1;
-	csinfo[i].fr_fp = NULL;
-	csinfo[i].to_fp = NULL;
+	csinfo[i].fname  = NULL;
+	csinfo[i].ppath  = NULL;
+	csinfo[i].flags  = NULL;
+	csinfo[i].pid    = -1;
+	csinfo[i].st_dev = NULL;
+	csinfo[i].st_ino = NULL;
+	csinfo[i].fr_fp  = NULL;
+	csinfo[i].to_fp  = NULL;
     }
 
     init_already = TRUE;
@@ -804,16 +808,25 @@ cs_init()
  * insert a new cscope database filename into the filelist
  */
 static int
-cs_insert_filelist(fname, ppath, flags)
+cs_insert_filelist(fname, ppath, flags, sb)
     char *fname;
     char *ppath;
     char *flags;
+    struct stat *sb;
 {
     short i;
 
-    for (i = 0; i < CSCOPE_MAX_CONNECTIONS; i++)
+    for (i = 0; i < CSCOPE_MAX_CONNECTIONS; i++) {
+	if (csinfo[i].fname &&
+	    csinfo[i].st_dev == sb->st_dev && csinfo[i].st_ino == sb->st_ino) {
+	    (void)EMSG("duplicate cscope database not added");
+	    return -1;
+	}
+
 	if (csinfo[i].fname == NULL)
 	    break;
+    }
+
     if (i == CSCOPE_MAX_CONNECTIONS) {
 	(void)EMSG("maximum number of cscope connections reached");
 	return -1;
@@ -846,6 +859,9 @@ cs_insert_filelist(fname, ppath, flags)
     } else
 	csinfo[i].flags = NULL;
 
+    csinfo[i].st_dev = sb->st_dev;
+    csinfo[i].st_ino = sb->st_ino;
+
     return i;
 } /* cs_insert_filelist */
 
@@ -887,26 +903,48 @@ static int
 cs_kill(eap)
     EXARG *eap;
 {
-    char *stok, printbuf[64];
+    char *stok, *printbuf = NULL;
     short i;
 
     if ((stok = strtok((char *)NULL, (const char *)" ")) == NULL) {
-	CS_USAGE_MSG(Find);
+	CS_USAGE_MSG(Kill);
 	return CSCOPE_FAILURE;
     }
 
-    i = atoi(stok);
-
-    if (i > CSCOPE_MAX_CONNECTIONS || i < 0 || csinfo[i].fname == NULL) {
-	(void)sprintf(printbuf, "cscope connection %s not found", stok);
-	(void)EMSG(printbuf);
+    if (strlen(stok) < 2 && isdigit(stok[0])) {
+	i = atoi(stok);
+	printbuf = (char *)alloc(strlen(stok) + 32);
     } else {
-	cs_release_csp(i, TRUE);
-	(void)sprintf(printbuf, "cscope connection %s closed", stok);
-	msg_clr_eos();
-	MSG_PUTS_ATTR(printbuf, highlight_attr[HLF_R]);
+	/* It must be part of a name.  We will try to find a match
+	 * within all the names in the csinfo data structure
+	 */
+	for (i = 0; i < CSCOPE_MAX_CONNECTIONS; i++) {
+	    if (csinfo[i].fname != NULL && strstr(csinfo[i].fname, stok)) {
+		printbuf = (char *)alloc(strlen(csinfo[i].fname) + 32);
+		break;
+	    }
+	}
     }
 
+    if (i >= CSCOPE_MAX_CONNECTIONS || i < 0 || csinfo[i].fname == NULL) {
+	if (printbuf == NULL)
+	    (void)EMSG("cscope connection not found");
+	else {
+	    sprintf(printbuf, "cscope connection %s not found", stok);
+	    (void)EMSG(printbuf);
+	}
+    } else {
+	if (printbuf == NULL)
+	    MSG_PUTS_ATTR("cscope connection closed", highlight_attr[HLF_R]);
+	else {
+	    cs_release_csp(i, TRUE);
+	    msg_clr_eos();
+	    sprintf(printbuf, "cscope connection %s closed", stok);
+	    MSG_PUTS_ATTR(printbuf, highlight_attr[HLF_R]);
+	}
+    }
+
+    vim_free(printbuf);
     return 0;
 } /* cs_kill */
 
@@ -1174,76 +1212,136 @@ cs_print_tags_priv(matches, cntxts, num_matches)
 {
     char buf[1024]; /* hope that 1024 is enough, else we need to malloc */
     char *fname, *lno, *extra, *tbuf;
-    char tname[64]; /* hope tag name length is < 64 chars */
-    int i;
+    int i, j, idx, num;
+    bool_t *in_cur_file;
 
     assert (num_matches > 0);
 
-    (void)sscanf(matches[0], "%s", tname);
-    (void)sprintf(buf, "Cscope tag: %s\n", tname);
+    if ((tbuf = (char *)alloc(strlen(matches[0]) + 1)) == NULL) {
+	MSG_PUTS_ATTR("couldn't malloc\n", highlight_attr[HLF_T]);
+	return;
+    }
+
+    strcpy(tbuf, matches[0]);
+    (void)sprintf(buf, "Cscope tag: %s\n", strtok(tbuf, "\t"));
+    vim_free(tbuf);
     MSG_PUTS_ATTR(buf, highlight_attr[HLF_T]);
 
     MSG_PUTS_ATTR("   #   line", highlight_attr[HLF_T]);    /* strlen is 7 */
     msg_advance(msg_col + 2);
     MSG_PUTS_ATTR("filename / context / line\n", highlight_attr[HLF_T]);
 
-    for (i = 0; i < num_matches; i++) {
-	/* if we really wanted to, we could avoid this malloc and strcpy
-	 * by parsing matches[i] on the fly and placing stuff into buf
-	 * directly, but that's too much of a hassle
-	 */
-	if ((tbuf = (char *)alloc(strlen(matches[i]) + 1)) == NULL)
-	    continue;
-	(void)strcpy(tbuf, matches[i]);
-
-	if ((fname = strtok(tbuf, (const char *)"\t")) == NULL)
-	    continue;
-	if ((fname = strtok(NULL, (const char *)"\t")) == NULL)
-	    continue;
-	if ((lno = strtok(NULL, (const char *)"\t")) == NULL)
+    /*
+     * for normal tags (non-cscope tags), vim sorts the tags before printing.
+     * hence, the output from 'matches' needs to be sorted as well (but we're
+     * not actually sorting the contents of 'matches').  for sorting, we simply
+     * print all tags in the current file before any other tags.  this idea
+     * was suggested by Jeffrey George <jgeorge@texax.net>
+     */
+    in_cur_file = (bool_t *)alloc_clear(num_matches * sizeof(bool_t));
+    if (in_cur_file != NULL)
+    {
+	if (curbuf->b_fname != NULL)
 	{
-	    /* if NULL, then no "extra", although in cscope's case, there
-	     * should always be "extra".
+	    char *f;
+	    int len;
+
+	    len = strlen((const char *)(curbuf->b_fname));
+	    for (i = 0; i < num_matches; i++)
+	    {
+		if ((f = strchr(matches[i], '\t')) != NULL)
+		{
+		    f++;
+		    if (strncmp((const char *)(curbuf->b_fname), f, len) == 0)
+			in_cur_file[i] = TRUE;
+		}
+	    }
+	}
+    }
+
+    /*
+     * if we were able to allocate 'in_cur_file', then we make two passes
+     * through 'matches'.  the first pass prints the tags in the current file,
+     * while the second prints all remaining tags.  if we were unable to
+     * allocate 'in_cur_file', we'll just make one pass through 'matches' and
+     * print them unsorted.
+     */
+    num = 1;
+    for (j = (in_cur_file) ? 2 : 1; j > 0; j--)
+    {
+	for (i = 0; i < num_matches; i++)
+	{
+	    if (in_cur_file)
+	    {
+		if (((j == 2) && (in_cur_file[i] == TRUE)) ||
+		    ((j == 1) && (in_cur_file[i] == FALSE)))
+			idx = i;
+		else
+		    continue;
+	    }
+	    else
+		idx = i;
+
+	    /* if we really wanted to, we could avoid this malloc and strcpy
+	     * by parsing matches[i] on the fly and placing stuff into buf
+	     * directly, but that's too much of a hassle
 	     */
-	     extra = NULL;
-	}
+	    if ((tbuf = (char *)alloc(strlen(matches[idx]) + 1)) == NULL)
+		continue;
+	    (void)strcpy(tbuf, matches[idx]);
 
-	extra = lno + strlen(lno) + 1;
+	    if ((fname = strtok(tbuf, (const char *)"\t")) == NULL)
+		continue;
+	    if ((fname = strtok(NULL, (const char *)"\t")) == NULL)
+		continue;
+	    if ((lno = strtok(NULL, (const char *)"\t")) == NULL)
+	    {
+		/* if NULL, then no "extra", although in cscope's case, there
+		 * should always be "extra".
+		 */
+		extra = NULL;
+	    }
 
-	lno[strlen(lno)-2] = '\0';  /* ignore ;" at the end */
+	    extra = lno + strlen(lno) + 1;
 
-	(void)sprintf(buf, "%4d %6s  ", i + 1, lno);
-	MSG_PUTS_ATTR(buf, highlight_attr[HLF_CM]);
-	MSG_PUTS_LONG_ATTR(fname, highlight_attr[HLF_CM]);
+	    lno[strlen(lno)-2] = '\0';  /* ignore ;" at the end */
 
-	(void)sprintf(buf, " <<%s>>",
-		(cntxts[i] == NULL) ? "GLOBAL" : cntxts[i]);
+	    (void)sprintf(buf, "%4d %6s  ", num, lno);
+	    MSG_PUTS_ATTR(buf, highlight_attr[HLF_CM]);
+	    MSG_PUTS_LONG_ATTR(fname, highlight_attr[HLF_CM]);
 
-	/* print the context only if it fits on the same line */
-	if (msg_col + strlen(buf) >= (int)Columns)
-	    msg_putchar('\n');
-	msg_advance(12);
-	MSG_PUTS_LONG(buf);
-	msg_putchar('\n');
+	    (void)sprintf(buf, " <<%s>>",
+			  (cntxts[idx] == NULL) ? "GLOBAL" : cntxts[idx]);
 
-	if (extra != NULL)
-	{
-	    msg_advance(13);
-	    MSG_PUTS_LONG(extra);
-	}
-
-	vim_free(tbuf); /* only free after printing extra due to strtok use */
-
-	if (msg_col)
+	    /* print the context only if it fits on the same line */
+	    if (msg_col + strlen(buf) >= (int)Columns)
+		msg_putchar('\n');
+	    msg_advance(12);
+	    MSG_PUTS_LONG(buf);
 	    msg_putchar('\n');
 
-	ui_breakcheck();
-	if (got_int) {
-	    got_int = FALSE;	/* don't print any more matches */
-	    break;
-	}
-    } /* for all matches */
+	    if (extra != NULL)
+	    {
+		msg_advance(13);
+		MSG_PUTS_LONG(extra);
+	    }
 
+	    vim_free(tbuf); /* only after printing extra due to strtok use */
+
+	    if (msg_col)
+		msg_putchar('\n');
+
+	    ui_breakcheck();
+	    if (got_int) {
+		got_int = FALSE;	/* don't print any more matches */
+		break;
+	    }
+
+	    num++;
+	} /* for all matches */
+    } /* 1 or 2 passes through 'matches' */
+
+    vim_free(in_cur_file);
 } /* cs_print_tags_priv */
 
 
@@ -1316,12 +1414,14 @@ cs_release_csp(i, freefnpp)
 	vim_free(csinfo[i].flags);
     }
 
-    csinfo[i].fname = NULL;
-    csinfo[i].ppath = NULL;
-    csinfo[i].flags = NULL;
-    csinfo[i].pid   = -1;
-    csinfo[i].fr_fp = NULL;
-    csinfo[i].to_fp = NULL;
+    csinfo[i].fname  = NULL;
+    csinfo[i].ppath  = NULL;
+    csinfo[i].flags  = NULL;
+    csinfo[i].pid    = -1;
+    csinfo[i].st_dev = NULL;
+    csinfo[i].st_ino = NULL;
+    csinfo[i].fr_fp  = NULL;
+    csinfo[i].to_fp  = NULL;
 } /* cs_release_csp */
 
 
