@@ -190,7 +190,9 @@ enum key_extra
     KE_XF1,		/* extra vt100 function keys for xterm */
     KE_XF2,
     KE_XF3,
-    KE_XF4
+    KE_XF4,
+    KE_XEND,		/* extra (vt100) end key for xterm */
+    KE_XHOME		/* extra (vt100) home key for xterm */
 };
 
 /*
@@ -303,12 +305,14 @@ enum key_extra
 #define K_INS		TERMCAP2KEY('k', 'I')
 #define K_DEL		TERMCAP2KEY('k', 'D')
 #define K_HOME		TERMCAP2KEY('k', 'h')
+#define K_KHOME		TERMCAP2KEY('K', '1')	/* keypad home (upper left) */
+#define K_XHOME		TERMCAP2KEY(KS_EXTRA, KE_XHOME)
 #define K_END		TERMCAP2KEY('@', '7')
+#define K_KEND		TERMCAP2KEY('K', '4')	/* keypad end (lower left) */
+#define K_XEND		TERMCAP2KEY(KS_EXTRA, KE_XEND)
 #define K_PAGEUP	TERMCAP2KEY('k', 'P')
 #define K_PAGEDOWN	TERMCAP2KEY('k', 'N')
-#define K_KHOME		TERMCAP2KEY('K', '1')	/* keypad home (upper left) */
 #define K_KPAGEUP	TERMCAP2KEY('K', '3')	/* keypad pageup (upper R.) */
-#define K_KEND		TERMCAP2KEY('K', '4')	/* keypad end (lower left) */
 #define K_KPAGEDOWN	TERMCAP2KEY('K', '5')	/* keypad pagedown (lower R.) */
 
 #define K_KPLUS		TERMCAP2KEY('K', '6')	/* keypad plus */
