@@ -14,6 +14,7 @@
 #endif
 
 static void mainerr __ARGS((int, char_u *));
+static void main_msg __ARGS((char *s));
 static void usage __ARGS((void));
 static int get_number_arg __ARGS((char_u *p, int *idx, int def));
 
@@ -58,6 +59,18 @@ mainerr(n, str)
     mch_windexit(1);
 }
 
+/*
+ * print a message with three spaces prepended and '\n' appended.
+ */
+    static void
+main_msg(s)
+    char *s;
+{
+    mch_msg("   ");
+    mch_msg(s);
+    mch_msg("\n");
+}
+
     static void
 usage()
 {
@@ -76,99 +89,111 @@ usage()
     reset_signals();		/* kill us with CTRL-C here, if you like */
 #endif
 
-    mch_errmsg(longVersion);
-    mch_errmsg("\nusage:");
+    mch_msg(longVersion);
+    mch_msg("\nusage:");
     for (i = 0; ; ++i)
     {
-	mch_errmsg(" vim [options] ");
-	mch_errmsg((char *)use[i]);
+	mch_msg(" vim [options] ");
+	mch_msg((char *)use[i]);
 	if (i == (sizeof(use) / sizeof(char_u *)) - 1)
 	    break;
-	mch_errmsg("\n   or:");
+	mch_msg("\n   or:");
     }
 
-    mch_errmsg("\n\nOptions:\n");
-    mch_errmsg("   --\t\t\tEnd of options\n");
+    mch_msg("\n\nOptions:\n");
+    main_msg("--\t\t\tEnd of options");
 #ifdef HAVE_OLE
-    mch_errmsg("   -register\t\tRegister this gvim for OLE\n");
-    mch_errmsg("   -unregister\t\tUnregister gvim for OLE\n");
+    main_msg("-register\t\tRegister this gvim for OLE");
+    main_msg("-unregister\t\tUnregister gvim for OLE");
 #endif
 #ifdef USE_GUI
-    mch_errmsg("   -g\t\t\tRun using GUI (like \"gvim\")\n");
-    mch_errmsg("   -f\t\t\tForeground: Don't fork when starting GUI\n");
+    main_msg("-g\t\t\tRun using GUI (like \"gvim\")");
+    main_msg("-f\t\t\tForeground: Don't fork when starting GUI");
 #endif
-    mch_errmsg("   -v\t\t\tVi mode (like \"vi\")\n");
-    mch_errmsg("   -e\t\t\tEx mode (like \"ex\")\n");
-    mch_errmsg("   -s\t\t\tSilent (batch) mode (only for \"ex\")\n");
-    mch_errmsg("   -R\t\t\tReadonly mode (like \"view\")\n");
-    mch_errmsg("   -Z\t\t\tRestricted mode (like \"rvim\")\n");
-    mch_errmsg("   -b\t\t\tBinary mode\n");
+    main_msg("-v\t\t\tVi mode (like \"vi\")");
+    main_msg("-e\t\t\tEx mode (like \"ex\")");
+    main_msg("-s\t\t\tSilent (batch) mode (only for \"ex\")");
+    main_msg("-R\t\t\tReadonly mode (like \"view\")");
+    main_msg("-Z\t\t\tRestricted mode (like \"rvim\")");
+    main_msg("-b\t\t\tBinary mode");
 #ifdef LISPINDENT
-    mch_errmsg("   -l\t\t\tLisp mode\n");
+    main_msg("-l\t\t\tLisp mode");
 #endif
-    mch_errmsg("   -C\t\t\tCompatible with Vi: 'compatible'\n");
-    mch_errmsg("   -N\t\t\tNot fully Vi compatible: 'nocompatible'\n");
-    mch_errmsg("   -V[N]\t\tVerbose level\n");
-    mch_errmsg("   -n\t\t\tNo swap file, use memory only\n");
-    mch_errmsg("   -r\t\t\tList swap files and exit\n");
-    mch_errmsg("   -r (with file name)\tRecover crashed session\n");
-    mch_errmsg("   -L\t\t\tSame as -r\n");
+    main_msg("-C\t\t\tCompatible with Vi: 'compatible'");
+    main_msg("-N\t\t\tNot fully Vi compatible: 'nocompatible'");
+    main_msg("-V[N]\t\tVerbose level");
+    main_msg("-n\t\t\tNo swap file, use memory only");
+    main_msg("-r\t\t\tList swap files and exit");
+    main_msg("-r (with file name)\tRecover crashed session");
+    main_msg("-L\t\t\tSame as -r");
 #ifdef AMIGA
-    mch_errmsg("   -f\t\t\tDon't use newcli to open window\n");
-    mch_errmsg("   -d <device>\t\tUse <device> for I/O\n");
+    main_msg("-f\t\t\tDon't use newcli to open window");
+    main_msg("-d <device>\t\tUse <device> for I/O");
 #endif
 #ifdef RIGHTLEFT
-    mch_errmsg("   -H\t\t\tstart in Hebrew mode\n");
+    main_msg("-H\t\t\tstart in Hebrew mode");
 #endif
 #ifdef FKMAP
-    mch_errmsg("   -F\t\t\tstart in Farsi mode\n");
+    main_msg("-F\t\t\tstart in Farsi mode");
 #endif
-    mch_errmsg("   -T <terminal>\tSet terminal type to <terminal>\n");
-    mch_errmsg("   -o[N]\t\tOpen N windows (default: one for each file)\n");
-    mch_errmsg("   +\t\t\tStart at end of file\n");
-    mch_errmsg("   +<lnum>\t\tStart at line <lnum>\n");
-    mch_errmsg("   -c <command>\t\tExecute <command> first\n");
-    mch_errmsg("   -s <scriptin>\tRead commands from script file <scriptin>\n");
-    mch_errmsg("   -w <scriptout>\tAppend commands to script file <scriptout>\n");
-    mch_errmsg("   -W <scriptout>\tWrite commands to script file <scriptout>\n");
-    mch_errmsg("   -u <vimrc>\t\tUse <vimrc> instead of any .vimrc\n");
+    main_msg("-T <terminal>\tSet terminal type to <terminal>");
+    main_msg("-o[N]\t\tOpen N windows (default: one for each file)");
+    main_msg("+\t\t\tStart at end of file");
+    main_msg("+<lnum>\t\tStart at line <lnum>");
+    main_msg("-c <command>\t\tExecute <command> first");
+    main_msg("-s <scriptin>\tRead commands from script file <scriptin>");
+    main_msg("-w <scriptout>\tAppend commands to script file <scriptout>");
+    main_msg("-W <scriptout>\tWrite commands to script file <scriptout>");
+    main_msg("-u <vimrc>\t\tUse <vimrc> instead of any .vimrc");
 #ifdef USE_GUI
-    mch_errmsg("   -U <gvimrc>\t\tUse <gvimrc> instead of any .gvimrc\n");
+    main_msg("-U <gvimrc>\t\tUse <gvimrc> instead of any .gvimrc");
+#endif
+#ifdef CRYPTV
+    main_msg("-x\t\t\tEdit encrypted files");
 #endif
 #ifdef VIMINFO
-    mch_errmsg("   -i <viminfo>\t\tUse <viminfo> instead of .viminfo\n");
+    main_msg("-i <viminfo>\t\tUse <viminfo> instead of .viminfo");
 #endif
-    mch_errmsg("   -h\t\t\tprint Help (this message) and exit\n");
-    mch_errmsg("   --version\t\tprint version information and exit\n");
+    main_msg("-h\t\t\tprint Help (this message) and exit");
+    main_msg("--version\t\tprint version information and exit");
 
 #ifdef USE_GUI_X11
 # ifdef USE_GUI_MOTIF
-    mch_errmsg("\nOptions recognised by gvim (Motif version):\n");
+    mch_msg("\nOptions recognised by gvim (Motif version):\n");
 # else
 #  ifdef USE_GUI_ATHENA
-    mch_errmsg("\nOptions recognised by gvim (Athena version):\n");
+    mch_msg("\nOptions recognised by gvim (Athena version):\n");
 #  endif /* USE_GUI_ATHENA */
 # endif /* USE_GUI_MOTIF */
-    mch_errmsg("   -display <display>\tRun vim on <display>\n");
-    mch_errmsg("   -iconic\t\tStart vim iconified\n");
+    main_msg("-display <display>\tRun vim on <display>");
+    main_msg("-iconic\t\tStart vim iconified");
 # if 0
-    mch_errmsg("   -name <name>\t\tUse resource as if vim was <name>\n");
-    mch_errmsg("\t\t\t  (Unimplemented)\n");
+    main_msg("-name <name>\t\tUse resource as if vim was <name>");
+    mch_msg("\t\t\t  (Unimplemented)\n");
 # endif
-    mch_errmsg("   -background <color>\tUse <color> for the background (also: -bg)\n");
-    mch_errmsg("   -foreground <color>\tUse <color> for normal text (also: -fg)\n");
-    mch_errmsg("   -font <font>\t\tUse <font> for normal text (also: -fn)\n");
-    mch_errmsg("   -boldfont <font>\tUse <font> for bold text\n");
-    mch_errmsg("   -italicfont <font>\tUse <font> for italic text\n");
-    mch_errmsg("   -geometry <geom>\tUse <geom> for initial geometry (also: -geom)\n");
-    mch_errmsg("   -borderwidth <width>\tUse a border width of <width> (also: -bw)\n");
-    mch_errmsg("   -scrollbarwidth <width>\tUse a scrollbar width of <width> (also: -sw)\n");
-    mch_errmsg("   -menuheight <height>\tUse a menu bar height of <height> (also: -mh)\n");
-    mch_errmsg("   -reverse\t\tUse reverse video (also: -rv)\n");
-    mch_errmsg("   +reverse\t\tDon't use reverse video (also: +rv)\n");
-    mch_errmsg("   -xrm <resource>\tSet the specified resource\n");
+    main_msg("-background <color>\tUse <color> for the background (also: -bg)");
+    main_msg("-foreground <color>\tUse <color> for normal text (also: -fg)");
+    main_msg("-font <font>\t\tUse <font> for normal text (also: -fn)");
+    main_msg("-boldfont <font>\tUse <font> for bold text");
+    main_msg("-italicfont <font>\tUse <font> for italic text");
+    main_msg("-geometry <geom>\tUse <geom> for initial geometry (also: -geom)");
+    main_msg("-borderwidth <width>\tUse a border width of <width> (also: -bw)");
+    main_msg("-scrollbarwidth <width>\tUse a scrollbar width of <width> (also: -sw)");
+    main_msg("-menuheight <height>\tUse a menu bar height of <height> (also: -mh)");
+    main_msg("-reverse\t\tUse reverse video (also: -rv)");
+    main_msg("+reverse\t\tDon't use reverse video (also: +rv)");
+    main_msg("-xrm <resource>\tSet the specified resource");
 #endif /* USE_GUI_X11 */
-
+#if defined(USE_GUI) && defined(RISCOS)
+    mch_msg("\nOptions recognised by gvim (RISC OS version):\n");
+    main_msg("--columns <number>\tInitial width of window in columns");
+    main_msg("--rows <number>\tInitial height of window in rows");
+#endif
+#ifdef USE_GUI_GTK
+   /* FIXME: describe them properly */
+   mch_msg("\nOptions recognised by gvim (GTK+ version):\n");
+   main_msg("...");
+#endif
     mch_windexit(1);
 }
 
@@ -195,6 +220,9 @@ main(argc, argv)
     char_u	   *use_vimrc = NULL;	    /* vimrc from -u option */
 #ifdef QUICKFIX
     char_u	   *use_ef = NULL;	    /* 'errorfile' from -q option */
+#endif
+#ifdef CRYPTV
+    int		    ask_for_key = FALSE;    /* -x argument */
 #endif
     int		    n_commands = 0;	    /* no. of commands from + or -c */
     char_u	   *commands[MAX_ARG_CMDS]; /* commands from + or -c option */
@@ -217,7 +245,11 @@ main(argc, argv)
     int		    stdout_isatty;	    /* is stdout a terminal? */
     int		    input_isatty;	    /* is active input a terminal? */
     OPARG	    oa;			    /* operator arguments */
-    WIN		    *wp;
+
+#ifdef RISCOS
+    /* Turn off all the horrible filename munging in UnixLib. */
+    __uname_control = __UNAME_NO_PROCESS;
+#endif
 
 #if defined(MSDOS) || defined(WIN32) || defined(OS2)
     /*
@@ -238,7 +270,7 @@ main(argc, argv)
 
 	/* paste, copy and cut */
 	{(char_u *)"<S-Insert> \"*P", NORMAL},
-	{(char_u *)"<S-Insert> \"\"d\"*P", VISUAL},
+	{(char_u *)"<S-Insert> \"-d\"*P", VISUAL},
 	{(char_u *)"<S-Insert> <C-R>*", INSERT+CMDLINE},
 	{(char_u *)"<C-Insert> \"*y", VISUAL},
 	{(char_u *)"<S-Del> \"*d", VISUAL},
@@ -258,12 +290,12 @@ main(argc, argv)
 	/* paste, copy and cut */
 #  ifdef USE_CLIPBOARD
 	{(char_u *)"\316\324 \"*P", NORMAL},	    /* SHIFT-Insert is "*P */
-	{(char_u *)"\316\324 \"\"d\"*P", VISUAL},   /* SHIFT-Insert is ""d"*P */
+	{(char_u *)"\316\324 \"-d\"*P", VISUAL},    /* SHIFT-Insert is "-d"*P */
 	{(char_u *)"\316\324 \017\"*P", INSERT},    /* SHIFT-Insert is ^O"*P */
 	{(char_u *)"\316\325 \"*y", VISUAL},	    /* CTRL-Insert is "*y */
 	{(char_u *)"\316\327 \"*d", VISUAL},	    /* SHIFT-Del is "*d */
 	{(char_u *)"\316\330 \"*d", VISUAL},	    /* CTRL-Del is "*d */
-	{(char_u *)"\030 \"\"d", VISUAL},	    /* CTRL-X is ""d */
+	{(char_u *)"\030 \"-d", VISUAL},	    /* CTRL-X is "-d */
 #  else
 	{(char_u *)"\316\324 P", NORMAL},	    /* SHIFT-Insert is P */
 	{(char_u *)"\316\324 d\"0P", VISUAL},	    /* SHIFT-Insert is d"0P */
@@ -289,14 +321,18 @@ main(argc, argv)
     {
 	/* paste, copy and cut */
 	{(char_u *)"<D-v> \"*P", NORMAL},
-	{(char_u *)"<D-v> \"\"d\"*P", VISUAL},
+	{(char_u *)"<D-v> \"-d\"*P", VISUAL},
 	{(char_u *)"<D-v> <C-R>*", INSERT+CMDLINE},
 	{(char_u *)"<D-c> \"*y", VISUAL},
 	{(char_u *)"<D-x> \"*d", VISUAL},
-	{(char_u *)"<Backspace> \"\"d", VISUAL},
+	{(char_u *)"<Backspace> \"-d", VISUAL},
     };
 #endif
 
+#ifdef MEM_PROFILE
+    atexit(vim_mem_profile_dump);
+#endif
+
 #ifdef __EMX__
     _wildcard(&argc, &argv);
 #endif
@@ -305,43 +341,6 @@ main(argc, argv)
     setlocale(LC_ALL, "");	/* for ctype() and the like */
 #endif
 
-#if defined(USE_GUI_WIN32) && defined(HAVE_OLE)
-    /* Check for special OLE command line parameters */
-    if (argc == 2 && (argv[1][0] == '-' || argv[1][0] == '/'))
-    {
-	/* Register Vim as an OLE Automation server */
-	if (STRICMP(argv[1] + 1, "register") == 0)
-	{
-	    RegisterMe();
-	    mch_windexit(0);
-	}
-
-	/* Unregister Vim as an OLE Automation server */
-	if (STRICMP(argv[1] + 1, "unregister") == 0)
-	{
-	    UnregisterMe(TRUE);
-	    mch_windexit(0);
-	}
-
-	/* Ignore an -embedding argument. It is only relevant if the
-	 * application wants to treat the case when it is started manually
-	 * differently from the case where it is started via automation (and
-	 * we don't).
-	 */
-	if (STRICMP(argv[1] + 1, "embedding") == 0)
-	    argc = 1;
-    }
-
-    {
-	int	bDoRestart = FALSE;
-
-	InitOLE(&bDoRestart);
-	/* automatically exit after registering */
-	if (bDoRestart)
-	    mch_windexit(0);
-    }
-#endif
-
 #ifdef USE_GUI
     gui_prepare(&argc, argv);	/* Prepare for possibly starting GUI sometime */
 #endif
@@ -475,11 +474,17 @@ main(argc, argv)
 	    c = argv[0][argv_idx++];
 	    switch (c)
 	    {
-	    case NUL:		/* "-"  read from stdin */
-		if (edit_type != EDIT_NONE)
-		    mainerr(ME_TOO_MANY_ARGS, (char_u *)argv[0]);
-		edit_type = EDIT_STDIN;
-		read_cmd_fd = 2;	/* read from stderr instead of stdin */
+	    case NUL:		/* "vim -"  read from stdin */
+				/* "ex -" silent mode */
+		if (exmode_active)
+		    silent_mode = TRUE;
+		else
+		{
+		    if (edit_type != EDIT_NONE)
+			mainerr(ME_TOO_MANY_ARGS, (char_u *)argv[0]);
+		    edit_type = EDIT_STDIN;
+		    read_cmd_fd = 2;	/* read from stderr instead of stdin */
+		}
 		argv_idx = -1;		/* skip to next argument */
 		break;
 
@@ -494,6 +499,8 @@ main(argc, argv)
 		    do_version((char_u *)"");
 		    mch_windexit(1);
 		}
+		if (argv[0][argv_idx])
+		    mainerr(ME_UNKNOWN_OPTION, (char_u *)argv[0]);
 		had_minmin = TRUE;
 		argv_idx = -1;		/* skip to next argument */
 		break;
@@ -633,9 +640,11 @@ main(argc, argv)
 		want_argument = TRUE;
 		break;
 
-	    case 'x':	    /* "-x"  use "crypt" for reading/writing files. */
-		/* TODO */
+#ifdef CRYPTV
+	    case 'x':	    /* "-x"  encrypted reading/writing of files */
+		ask_for_key = TRUE;
 		break;
+#endif
 
 	    case 'Z':		/* "-Z"  restricted mode */
 		restricted = TRUE;
@@ -907,7 +916,7 @@ main(argc, argv)
 	    initstr = vim_strsave(initmappings[i].arg);
 	    if (initstr != NULL)
 	    {
-		do_map(0, initstr, initmappings[i].mode, FALSE);
+		do_map(0, initstr, initmappings[i].mode, FALSE, NULL);
 		vim_free(initstr);
 	    }
 	}
@@ -916,7 +925,7 @@ main(argc, argv)
 #endif
 
     init_highlight(TRUE);	/* set the default highlight groups */
-#ifdef USE_GUI
+#ifdef CURSOR_SHAPE
     parse_guicursor();		/* set cursor shapes from 'guicursor' */
 #endif
 
@@ -963,6 +972,9 @@ main(argc, argv)
 	    if (do_source((char_u *)USR_VIMRC_FILE, TRUE, TRUE) == FAIL
 #ifdef USR_VIMRC_FILE2
 		&& do_source((char_u *)USR_VIMRC_FILE2, TRUE, TRUE) == FAIL
+#endif
+#ifdef USR_VIMRC_FILE3
+		&& do_source((char_u *)USR_VIMRC_FILE3, TRUE, TRUE) == FAIL
 #endif
 		&& process_env((char_u *)"EXINIT") == FAIL
 		&& do_source((char_u *)USR_EXRC_FILE, FALSE, FALSE) == FAIL)
@@ -1003,6 +1015,10 @@ main(argc, argv)
 		    && fullpathcmp((char_u *)USR_VIMRC_FILE2,
 				      (char_u *)VIMRC_FILE, FALSE) != FPC_SAME
 #endif
+#ifdef USR_VIMRC_FILE3
+		    && fullpathcmp((char_u *)USR_VIMRC_FILE3,
+				      (char_u *)VIMRC_FILE, FALSE) != FPC_SAME
+#endif
 #ifdef SYS_VIMRC_FILE
 		    && fullpathcmp((char_u *)SYS_VIMRC_FILE,
 				      (char_u *)VIMRC_FILE, FALSE) != FPC_SAME
@@ -1100,7 +1116,7 @@ main(argc, argv)
     {
 	if (use_ef != NULL)
 	    set_string_option_direct((char_u *)"ef", -1, use_ef, TRUE);
-	if (qf_init(p_ef) < 0)
+	if (qf_init(p_ef, p_efm) < 0)
 	{
 	    out_char('\n');
 	    mch_windexit(3);
@@ -1171,6 +1187,7 @@ main(argc, argv)
 	scroll_region_reset();		/* In case Rows changed */
 
     scroll_start();
+
     /*
      * Don't clear the screen when starting in Ex mode, unless using the GUI.
      */
@@ -1183,6 +1200,11 @@ main(argc, argv)
     else
 	screenclear();			/* clear screen */
 
+#ifdef CRYPTV
+    if (ask_for_key)
+	get_crypt_key(TRUE);
+#endif
+
     no_wait_return = TRUE;
 
     if (recoverymode)			/* do recover */
@@ -1255,7 +1277,7 @@ main(argc, argv)
 	curwin->w_cursor.lnum = curbuf->b_ml.ml_line_count;
 
 #ifdef AUTOCMD
-    apply_autocmds(EVENT_BUFENTER, NULL, NULL, FALSE);
+    apply_autocmds(EVENT_BUFENTER, NULL, NULL, FALSE, curbuf);
 #endif
     setpcmark();
 
@@ -1368,7 +1390,11 @@ main(argc, argv)
 	need_start_insertmode = TRUE;
 
 #ifdef AUTOCMD
-    apply_autocmds(EVENT_VIMENTER, NULL, NULL, FALSE);
+    apply_autocmds(EVENT_VIMENTER, NULL, NULL, FALSE, curbuf);
+#endif
+
+#if defined(WIN32) && !defined(USE_GUI_WIN32)
+    mch_set_winsize_now();	    /* Allow winsize changes from now on */
 #endif
 
     /*
@@ -1383,7 +1409,7 @@ main(argc, argv)
 		check_timestamps();
 	    if (need_wait_return)	/* if wait_return still needed ... */
 		wait_return(FALSE);	/* ... call it now */
-	    if (need_start_insertmode)
+	    if (need_start_insertmode && goto_im())
 	    {
 		need_start_insertmode = FALSE;
 		stuffReadbuff((char_u *)"i");	/* start insert mode next */
@@ -1412,11 +1438,10 @@ main(argc, argv)
 	{
 	    /*
 	     * Before redrawing, make sure w_topline is correct, and w_leftcol
-	     * if lines don't wrap.
+	     * if lines don't wrap, and w_skipcol if lines wrap.
 	     */
 	    update_topline();
-	    if (!curwin->w_p_wrap)
-		validate_cursor();
+	    validate_cursor();
 
 	    if (VIsual_active)
 		update_curbuf(INVERTED);/* update inverted part */
@@ -1424,9 +1449,7 @@ main(argc, argv)
 		update_screen(must_redraw);
 	    else if (redraw_cmdline || clear_cmdline)
 		showmode();
-	    for (wp = firstwin; wp; wp = wp->w_next)
-		if (wp->w_redr_status)
-		    win_redr_status(wp);
+	    redraw_statuslines();
 	    /* display message after redraw */
 	    if (keep_msg != NULL)
 		msg_attr(keep_msg, keep_msg_attr);
@@ -1444,6 +1467,10 @@ main(argc, argv)
 	    setcursor();
 	    cursor_on();
 	}
+#ifdef USE_GUI
+	if (need_mouse_correct)
+	    gui_mouse_correct();
+#endif
 
 	/*
 	 * Update w_curswant if w_set_curswant has been set.
@@ -1494,12 +1521,14 @@ process_env(env)
     char_u	*env;
 {
     char_u	*initstr;
+    char_u	*save_sourcing_name;
 
-    if ((initstr = vim_getenv(env)) != NULL && *initstr != NUL)
+    if ((initstr = mch_getenv(env)) != NULL && *initstr != NUL)
     {
+	save_sourcing_name = sourcing_name;
 	sourcing_name = env;
 	do_cmdline(initstr, NULL, NULL, DOCMD_NOWAIT|DOCMD_VERBOSE);
-	sourcing_name = NULL;
+	sourcing_name = save_sourcing_name;
 	return OK;
     }
     return FAIL;
@@ -1517,6 +1546,10 @@ getout(r)
 #endif
 	windgoto((int)Rows - 1, 0);
 
+#ifdef AUTOCMD
+    apply_autocmds(EVENT_VIMLEAVEPRE, NULL, NULL, FALSE, curbuf);
+#endif
+
 #ifdef VIMINFO
     if (*p_viminfo != NUL)
     {
@@ -1532,11 +1565,7 @@ getout(r)
 #endif /* VIMINFO */
 
 #ifdef AUTOCMD
-    apply_autocmds(EVENT_VIMLEAVE, NULL, NULL, FALSE);
-
-#ifdef HAVE_PERL_INTERP
-    perl_end();
-#endif
+    apply_autocmds(EVENT_VIMLEAVE, NULL, NULL, FALSE, curbuf);
 
     /* Position the cursor again, the autocommands may have moved it */
 # ifdef USE_GUI
@@ -1545,12 +1574,16 @@ getout(r)
 	windgoto((int)Rows - 1, 0);
 #endif
 
+#ifdef HAVE_PERL_INTERP
+    perl_end();
+#endif
+
     mch_windexit(r);
 }
 
 /*
  * When FKMAP is defined, also compile the Farsi source code.
  */
-#ifdef FKMAP
+#if defined(FKMAP) || defined(PROTO)
 # include "farsi.c"
 #endif
