@@ -606,19 +606,19 @@ msg_make(arg)
  * Return the number of characters it takes on the screen.
  *
  * If K_SPECIAL is encountered, then it is taken in conjunction with the
- * following character and shown as <F1>, <S-Up> etc.  In addition, if 'all'
- * is TRUE, then any other character which has its 8th bit set is shown as
- * <M-x>, where x is the equivalent character without its 8th bit set.	If a
- * character is displayed in one of these special ways, is also highlighted
- * (its highlight name is '8' in the p_hl variable).
+ * following character and shown as <F1>, <S-Up> etc.  Any other character
+ * which is not printable shown in <> form.
+ * If 'from' is TRUE (lhs of a mapping), a space is shown as <Space>.
+ * If a character is displayed in one of these special ways, is also
+ * highlighted (its highlight name is '8' in the p_hl variable).
  * Otherwise characters are not highlighted.
  * This function is used to show mappings, where we want to see how to type
  * the character/string -- webb
  */
     int
-msg_outtrans_special(str, all)
+msg_outtrans_special(str, from)
     char_u	*str;
-    int		all;	/* <M-a> etc as well as <F1> etc */
+    int		from;	/* lhs of a mapping */
 {
     int	    retval = 0;
     char_u  *string;
@@ -654,9 +654,9 @@ msg_outtrans_special(str, all)
 		continue;
 	    }
 	}
-	/* output unprintable meta characters, and <M-Space> */
-	if (all && (((c & 0x80) && (!vim_isprintc(c) || c == 0xa0))
-		    || c == ' '))
+	/* Output unprintable characters in <> form, also <M-Space> and <Tab>.
+	 * Output <Space> only for lhs of a mapping. */
+	if (charsize(c) > 1 || (from && c == ' '))
 	{
 	    string = get_special_key_name(c, 0);
 	    msg_puts_attr(string, attr);
@@ -664,8 +664,8 @@ msg_outtrans_special(str, all)
 	}
 	else
 	{
-	    msg_puts(transchar(c));
-	    retval += charsize(c);
+	    msg_putchar(c);
+	    ++retval;
 	}
     }
     return retval;
