@@ -25,6 +25,14 @@
 
 #include "vim.h"
 
+#ifdef USE_FONTSET
+# ifdef X_LOCALE
+#  include <X11/Xlocale.h>
+# else
+#  include <locale.h>
+# endif
+#endif
+
 #ifdef HAVE_FCNTL_H
 # include <fcntl.h>
 #endif
@@ -43,6 +51,27 @@
 extern int   select __ARGS((int, fd_set *, fd_set *, fd_set *, struct timeval *));
 #endif
 
+#ifdef GPM_MOUSE
+# include <gpm.h>
+/* <linux/keyboard.h> contains defines conflicting with "keymap.h",
+ * I just copied relevant defines here. A cleaner solution would be to put gpm
+ * code into separate file and include there linux/keyboard.h
+ */
+/* #include <linux/keyboard.h> */
+# define KG_SHIFT	0
+# define KG_CTRL	2
+# define KG_ALT		3
+# define KG_ALTGR	1
+# define KG_SHIFTL	4
+# define KG_SHIFTR	5
+# define KG_CTRLL	6
+# define KG_CTRLR	7
+# define KG_CAPSSHIFT	8
+
+static void gpm_close __ARGS((void));
+static int gpm_open __ARGS((void));
+#endif
+
 /*
  * end of autoconf section. To be extended...
  */
@@ -83,9 +112,9 @@ static void may_core_dump __ARGS((void));
 
 static int  WaitForChar __ARGS((long));
 #ifdef __BEOS__
-int  RealWaitForChar __ARGS((int, long));
+int  RealWaitForChar __ARGS((int, long, int *));
 #else
-static int  RealWaitForChar __ARGS((int, long));
+static int  RealWaitForChar __ARGS((int, long, int *));
 #endif
 
 #if defined(SIGWINCH)
@@ -94,7 +123,8 @@ static RETSIGTYPE sig_winch __ARGS(SIGPROTOARG);
 #if defined(SIGINT)
 static RETSIGTYPE catch_sigint __ARGS(SIGPROTOARG);
 #endif
-#if defined(SIGALRM) && defined(HAVE_X11) && defined(WANT_X11)
+#if defined(SIGALRM) && defined(HAVE_X11) && defined(WANT_X11) && !defined(USE_GUI_GTK)
+# define SET_SIG_ALARM
 static RETSIGTYPE sig_alarm __ARGS(SIGPROTOARG);
 #endif
 static RETSIGTYPE deathtrap __ARGS(SIGPROTOARG);
@@ -225,7 +255,7 @@ mch_write(s, len)
 {
     write(1, (char *)s, len);
     if (p_wd)		/* Unix is too fast, slow down a bit more */
-	RealWaitForChar(read_cmd_fd, p_wd);
+	RealWaitForChar(read_cmd_fd, p_wd, NULL);
 }
 
 /*
@@ -339,6 +369,10 @@ mch_delay(msec, ignoreinput)
 
 	tv.tv_sec = msec / 1000;
 	tv.tv_usec = (msec % 1000) * 1000;
+	/*
+	 * NOTE: Solaris 2.6 has a bug that makes select() hang here.  Get a
+	 * patch from Sun to fix this.  Reported by Gunnar Pedersen.
+	 */
 	select(0, NULL, NULL, NULL, &tv);
     }
 # endif	/* __EMX__ */
@@ -376,7 +410,7 @@ catch_sigint SIGDEFARG(sigarg)
 }
 #endif
 
-#if defined(SIGALRM) && defined(HAVE_X11) && defined(WANT_X11)
+#ifdef SET_SIG_ALARM
 /*
  * signal function for alarm().
  */
@@ -615,6 +649,11 @@ x_error_check(dpy, error_event)
  * try to get x11 window and display
  *
  * return FAIL for failure, OK otherwise
+ *
+ * FIXME:
+ * This code is responsible for setting the window manager name of the
+ * editor's window. We should provide a special version of this for usage with
+ * the GTK+ widget set, which wouldn't use any direct x11 calls. (--mdcki)
  */
     static int
 get_x11_windis()
@@ -628,7 +667,7 @@ get_x11_windis()
     /* X just exits if it finds an error otherwise! */
     XSetErrorHandler(x_error_handler);
 
-#ifdef USE_GUI_X11
+#if defined(USE_GUI_X11) || defined(USE_GUI_GTK)
     if (gui.in_use)
     {
 	/*
@@ -658,7 +697,7 @@ get_x11_windis()
 	x11_window = (Window)atol(winid);
     if (x11_window != 0 && x11_display == NULL)
     {
-#ifdef SIGALRM
+#ifdef SET_SIG_ALARM
 	RETSIGTYPE (*sig_save)();
 
 	/*
@@ -670,7 +709,12 @@ get_x11_windis()
 	alarm(2);
 #endif
 	x11_display = XOpenDisplay(NULL);
-#ifdef SIGALRM
+
+#ifdef USE_FONTSET
+	setlocale(LC_ALL, "");
+#endif
+
+#ifdef SET_SIG_ALARM
 	alarm(0);
 	signal(SIGALRM, (RETSIGTYPE (*)())sig_save);
 #endif
@@ -799,10 +843,42 @@ get_x11_thing(get_title, test_only)
 	    retval = TRUE;
 	    if (!test_only)
 	    {
-		if (get_title)
-		    oldtitle = vim_strsave((char_u *)text_prop.value);
+#ifdef USE_FONTSET
+		if (text_prop.encoding == XA_STRING)
+		{
+#endif
+		    if (get_title)
+			oldtitle = vim_strsave((char_u *)text_prop.value);
+		    else
+			oldicon = vim_strsave((char_u *)text_prop.value);
+#ifdef USE_FONTSET
+		}
 		else
-		    oldicon = vim_strsave((char_u *)text_prop.value);
+		{
+		    char    **cl;
+		    Status  transform_status;
+		    int	    n = 0;
+
+		    transform_status = XmbTextPropertyToTextList(x11_display,
+								 &text_prop,
+								 &cl, &n);
+		    if (transform_status >= Success && n > 0 && cl[0])
+		    {
+			if (get_title)
+			    oldtitle = vim_strsave((char_u *) cl[0]);
+			else
+			    oldicon = vim_strsave((char_u *) cl[0]);
+			XFreeStringList(cl);
+		    }
+		    else
+		    {
+			if (get_title)
+			    oldtitle = vim_strsave((char_u *)text_prop.value);
+			else
+			    oldicon = vim_strsave((char_u *)text_prop.value);
+		    }
+		}
+#endif
 	    }
 	    XFree((void *)text_prop.value);
 	}
@@ -1595,6 +1671,21 @@ mch_setmouse(on)
 	    out_str_nf((char_u *)"\033[?1000l"); /* disable mouse events */
 	ison = on;
     }
+# ifdef GPM_MOUSE
+    else
+    {
+        if (on)
+        {
+            if (gpm_open())
+                ison = 1;
+        }
+        else
+        {
+            gpm_close();
+            ison = 0;
+        }
+    }
+# endif
 }
 #endif
 
@@ -2235,7 +2326,7 @@ mch_call_shell(cmd, options)
 		     * TODO: This should handle escape sequences, compatible
 		     * to some terminal (vt52?).
 		     */
-		    while (RealWaitForChar(fromshell_fd, 10L))
+		    while (RealWaitForChar(fromshell_fd, 10L, NULL))
 		    {
 			len = read(fromshell_fd, (char *)buffer,
 							      (size_t)BUFLEN);
@@ -2340,7 +2431,7 @@ error:
     void
 mch_breakcheck()
 {
-    if (curr_tmode == TMODE_RAW && RealWaitForChar(read_cmd_fd, 0L))
+    if (curr_tmode == TMODE_RAW && RealWaitForChar(read_cmd_fd, 0L, NULL))
 	fill_input_buf(FALSE);
 }
 
@@ -2353,6 +2444,10 @@ mch_breakcheck()
 WaitForChar(msec)
     long    msec;
 {
+#ifdef GPM_MOUSE
+    int gpm_process_wanted;
+#endif
+
     if (!vim_is_input_buf_empty())	    /* something in inbuf[] */
 	return 1;
 
@@ -2365,7 +2460,19 @@ WaitForChar(msec)
     }
 #endif
 
-    return RealWaitForChar(read_cmd_fd, msec);
+#ifdef GPM_MOUSE
+    do
+    {
+        gpm_process_wanted = 0;
+        if (!RealWaitForChar(read_cmd_fd, msec, &gpm_process_wanted))
+            return  0;
+    }
+    while (gpm_process_wanted && mch_gpm_process() == 0);
+
+    return 1;
+#else
+    return RealWaitForChar(read_cmd_fd, msec, NULL);
+#endif
 }
 
 /*
@@ -2373,15 +2480,18 @@ WaitForChar(msec)
  * Time == -1 will block forever.
  * When a GUI is being used, this will not be used for input -- webb
  * Returns also, when a request from Sniff is waiting -- toni.
+ * Or when a Linux GPM mouse event is waiting.
  */
+/* ARGSUSED */
 #ifdef __BEOS__
     int
 #else
     static  int
 #endif
-RealWaitForChar(fd, msec)
+RealWaitForChar(fd, msec, check_for_gpm)
     int	    fd;
     long    msec;
+    int     *check_for_gpm;
 {
 #ifndef HAVE_SELECT
 # ifdef USE_SNIFF
@@ -2461,8 +2571,27 @@ RealWaitForChar(fd, msec)
     }
 # endif
 
+#ifdef GPM_MOUSE
+    if (check_for_gpm && gpm_flag && gpm_fd >= 0)
+    {
+        FD_SET(gpm_fd, &rfds);
+        if (maxfd < gpm_fd)
+            maxfd = gpm_fd;
+    }
+#endif
+
     ret = select(maxfd + 1, &rfds, NULL, &efds, (msec >= 0) ? &tv : NULL);
 
+#ifdef GPM_MOUSE
+    if (gpm_flag && check_for_gpm)
+    {
+        if (ret > 0 && FD_ISSET(gpm_fd, &efds))
+            gpm_close();
+        else if (ret > 0 && FD_ISSET(gpm_fd, &rfds))
+            *check_for_gpm = 1;
+    }
+#endif
+
 # ifdef USE_SNIFF
     if (ret < 0 )
 	sniff_disconnect(1);
@@ -2475,6 +2604,7 @@ RealWaitForChar(fd, msec)
     }
 # endif
 
+
     return (ret > 0);
 #endif
 }
@@ -3123,3 +3253,133 @@ mch_rename(src, dest)
     return -1;
 }
 #endif /* !HAVE_RENAME */
+
+#ifdef GPM_MOUSE
+/*
+ * Initializes connection with gpm (if it isn't already opened)
+ * Return 1 if succeeded (or connection already opened), 0 if failed
+ */
+    static int
+gpm_open()
+{
+    /* mouse termcode */
+    static char mouse_string[] = {ESC,'M','G','\0'};
+    static Gpm_Connect gpm_connect; /* Must it be kept till closing ? */
+    if (!gpm_flag)
+    {
+        gpm_connect.eventMask = (GPM_UP | GPM_DRAG | GPM_DOWN);
+        gpm_connect.defaultMask = ~GPM_HARD;
+        /* Default handling for mouse move*/
+        gpm_connect.minMod = 0; /* Handle any modifier keys */
+        gpm_connect.maxMod = 0xffff;
+        if (Gpm_Open(&gpm_connect, 0) > 0)
+        {
+            set_mouse_termcode(KS_MOUSE, mouse_string);
+            /* gpm library tries to handling TSTP causes
+             * problems. Anyways, we close connection to Gpm whenever
+             * we are going to suspend or starting an external process
+             * so we should'nt  have problem with this
+             */
+            signal(SIGTSTP, restricted ? SIG_IGN : SIG_DFL);
+            return 1; /* succeed */
+        }
+        if (gpm_fd == -2)
+            Gpm_Close(); /* We don't want to talk to xterm via gpm */
+        return 0;
+    }
+    return 1; /* already open */
+}
+
+/*
+ * Closes connection to gpm
+ * returns non-zero if connection succesfully closed
+ */
+    static void
+gpm_close()
+{
+    if (gpm_flag) /* Open */
+    {
+        Gpm_Close();
+        del_mouse_termcode(KS_MOUSE);
+    }
+}
+
+/* Reads gpm event and adds special keys to input buf. Returns length of
+ * generated key sequence.
+ * This function is made after gui_send_mouse_event
+ */
+    int
+mch_gpm_process()
+{
+    int			button;
+    static Gpm_Event	gpm_event;
+    char_u		string[6];
+    int_u		vim_modifiers;
+    int			row,col;
+    unsigned char	buttons_mask;
+    unsigned char	gpm_modifiers;
+    static unsigned char old_buttons = 0;
+
+    Gpm_GetEvent(&gpm_event);
+
+    row = gpm_event.y - 1;
+    col = gpm_event.x - 1;
+
+    string[0] = ESC; /* Our termcode */
+    string[1] = 'M';
+    string[2] = 'G';
+    switch (GPM_BARE_EVENTS(gpm_event.type))
+    {
+        case GPM_DRAG:
+            string[3] = MOUSE_DRAG;
+            break;
+        case GPM_DOWN:
+            buttons_mask = gpm_event.buttons & ~old_buttons;
+            old_buttons = gpm_event.buttons;
+            switch (buttons_mask)
+            {
+                case GPM_B_LEFT:
+                    button = MOUSE_LEFT;
+                    break;
+                case GPM_B_MIDDLE:
+                    button = MOUSE_MIDDLE;
+                    break;
+                case GPM_B_RIGHT:
+                    button = MOUSE_RIGHT;
+                    break;
+                default:
+                    return 0;
+                    /*Don't know what to do. Can more than one button be
+                     * reported in one event? */
+            }
+            string[3] = (char_u)(button | 0x20);
+            SET_NUM_MOUSE_CLICKS(string[3], gpm_event.clicks + 1);
+            break;
+        case GPM_UP:
+            string[3] = MOUSE_RELEASE;
+            old_buttons &= ~gpm_event.buttons;
+            break;
+        default:
+            return 0;
+    }
+    /*This code is based on gui_x11_mouse_cb in gui_x11.c */
+    gpm_modifiers = gpm_event.modifiers;
+    vim_modifiers = 0x0;
+    /* I ignore capslock stats. Aren't we all just hate capslock mixing with
+     * Vim commands ? Besides, gpm_event.modifiers is unsigned char, and
+     * K_CAPSSHIFT is defined 8, so it probably isn't even reported
+     */
+    if (gpm_modifiers & ((1 << KG_SHIFT) | (1 << KG_SHIFTR) | (1 << KG_SHIFTL)))
+	vim_modifiers |= MOUSE_SHIFT;
+
+    if (gpm_modifiers & ((1 << KG_CTRL) | (1 << KG_CTRLR) | (1 << KG_CTRLL)))
+	vim_modifiers |= MOUSE_CTRL;
+    if (gpm_modifiers & ((1 << KG_ALT) | (1 << KG_ALTGR)))
+	vim_modifiers |= MOUSE_ALT;
+    string[3] |= vim_modifiers;
+    string[4] = (char_u)(col + ' ' + 1);
+    string[5] = (char_u)(row + ' ' + 1);
+    add_to_input_buf(string, 6);
+    return 6;
+}
+#endif /* GPM_MOUSE */
