@@ -633,7 +633,10 @@ do_search(oap, dirc, str, count, options)
     }
 
 #ifdef EXTRA_SEARCH
-    if (no_hlsearch)
+    /*
+     * Turn 'hlsearch' highlighting back on.
+     */
+    if (no_hlsearch && !(options & SEARCH_KEEP))
     {
 	redraw_all_later(NOT_VALID);
 	no_hlsearch = FALSE;
@@ -1533,31 +1536,44 @@ findmatchlimit(oap, initc, flags, maxtravel)
 	 *   Skip things in single quotes: 'x' or '\x'.  Be careful for single
 	 *   single quotes, eg jon's.  Things like '\233' or '\x3f' are not
 	 *   skipped, there is never a brace in them.
+	 *   Ignore this when finding matches for `'.
 	 */
 	case '\'':
-	    if (!cpo_match)
+	    if (!cpo_match && initc != '\'' && findc != '\'')
 	    {
 		if (backwards)
 		{
 		    if (pos.col > 1)
 		    {
 			if (linep[pos.col - 2] == '\'')
+			{
 			    pos.col -= 2;
+			    break;
+			}
 			else if (linep[pos.col - 2] == '\\' &&
 				    pos.col > 2 && linep[pos.col - 3] == '\'')
+			{
 			    pos.col -= 3;
+			    break;
+			}
 		    }
 		}
 		else if (linep[pos.col + 1])	/* forward search */
 		{
 		    if (linep[pos.col + 1] == '\\' &&
 			    linep[pos.col + 2] && linep[pos.col + 3] == '\'')
+		    {
 			pos.col += 3;
+			break;
+		    }
 		    else if (linep[pos.col + 2] == '\'')
+		    {
 			pos.col += 2;
+			break;
+		    }
 		}
 	    }
-	    break;
+	    /* FALLTHROUGH */
 
 	default:
 	    /* Check for match outside of quotes, and inside of
@@ -1695,12 +1711,12 @@ findsent(dir, count)
 	/*
 	 * if on an empty line, skip upto a non-empty line
 	 */
-	if (gchar(&pos) == NUL)
+	if (gchar_pos(&pos) == NUL)
 	{
 	    do
 		if ((*func)(&pos) == -1)
 		    break;
-	    while (gchar(&pos) == NUL);
+	    while (gchar_pos(&pos) == NUL);
 	    if (dir == FORWARD)
 		goto found;
 	}
@@ -1720,7 +1736,7 @@ findsent(dir, count)
 	    decl(&pos);
 
 	/* go back to the previous non-blank char */
-	while ((c = gchar(&pos)) == ' ' || c == '\t' ||
+	while ((c = gchar_pos(&pos)) == ' ' || c == '\t' ||
 	     (dir == BACKWARD && vim_strchr((char_u *)".!?)]\"'", c) != NULL))
 	{
 	    if (decl(&pos) == -1)
@@ -1738,7 +1754,7 @@ findsent(dir, count)
 
 	for (;;)		/* find end of sentence */
 	{
-	    c = gchar(&pos);
+	    c = gchar_pos(&pos);
 	    if (c == NUL || (pos.col == 0 && startPS(pos.lnum, NUL, FALSE)))
 	    {
 		if (dir == BACKWARD && pos.lnum != startlnum)
@@ -1751,11 +1767,11 @@ findsent(dir, count)
 		do
 		    if ((c = inc(&tpos)) == -1)
 			break;
-		while (vim_strchr((char_u *)")]\"'", c = gchar(&tpos)) != NULL);
+		while (vim_strchr((char_u *)")]\"'", c = gchar_pos(&tpos)) != NULL);
 		if (c == -1  || c == ' ' || c == '\t' || c == NUL)
 		{
 		    pos = tpos;
-		    if (gchar(&pos) == NUL) /* skip NUL at EOL */
+		    if (gchar_pos(&pos) == NUL) /* skip NUL at EOL */
 			inc(&pos);
 		    break;
 		}
@@ -1770,7 +1786,7 @@ findsent(dir, count)
 	}
 found:
 	    /* skip white space */
-	while (!noskip && ((c = gchar(&pos)) == ' ' || c == '\t'))
+	while (!noskip && ((c = gchar_pos(&pos)) == ' ' || c == '\t'))
 	    if (incl(&pos) == -1)
 		break;
     }
@@ -2213,7 +2229,7 @@ find_first_blank(posp)
 
     while (decl(posp) != -1)
     {
-	c = gchar(posp);
+	c = gchar_pos(posp);
 	if (!vim_iswhite(c))
 	{
 	    incl(posp);
@@ -2439,7 +2455,7 @@ extend:
 	    decl(&pos);
 	    while (lt(pos, curwin->w_cursor))
 	    {
-		c = gchar(&pos);
+		c = gchar_pos(&pos);
 		if (!vim_iswhite(c))
 		{
 		    at_start_sent = FALSE;
@@ -2484,7 +2500,7 @@ extend:
 		at_start_sent = FALSE;
 		while (lt(pos, curwin->w_cursor))
 		{
-		    c = gchar(&pos);
+		    c = gchar_pos(&pos);
 		    if (!vim_iswhite(c))
 		    {
 			at_start_sent = TRUE;
@@ -2509,7 +2525,7 @@ extend:
      * If cursor started on blank, check if it is just before the start of the
      * next sentence.
      */
-    while (c = gchar(&pos), vim_iswhite(c))	/* vim_iswhite() is a macro */
+    while (c = gchar_pos(&pos), vim_iswhite(c))	/* vim_iswhite() is a macro */
 	incl(&pos);
     if (equal(pos, curwin->w_cursor))
     {
