@@ -642,7 +642,11 @@ struct builtin_term builtin_termcaps[] =
     {(int)KS_VB,	"\033|B"},	/* visual bell */
     {(int)KS_TI,	"\033|S"},	/* put terminal in termcap mode */
     {(int)KS_TE,	"\033|E"},	/* out of termcap mode */
+#  ifdef TERMINFO
+    {(int)KS_CS,	"\033|%i%p1%d;%p2%dr"},/* scroll region */
+#  else
     {(int)KS_CS,	"\033|%i%d;%dr"},/* scroll region */
+#  endif
 
     {K_UP,		"\316H"},
     {K_DOWN,		"\316P"},
@@ -825,8 +829,10 @@ struct builtin_term builtin_termcaps[] =
     {K_INS,		"\033[2~"},
     {K_HOME,		"\033[7~"},
     {K_KHOME,		"\033[1~"},
+    {K_XHOME,		"\033OH"},	/* alternate Home */
     {K_END,		"\033[8~"},
     {K_KEND,		"\033[4~"},
+    {K_XEND,		"\033OF"},	/* alternate End */
     {K_PAGEUP,		"\033[5~"},
     {K_PAGEDOWN,	"\033[6~"},
     {K_KPLUS,		"\033Ok"},	/* keypad plus */
@@ -865,25 +871,32 @@ struct builtin_term builtin_termcaps[] =
 #  else
     {(int)KS_CDL,	"\033[%dM"},
 #  endif
-/*
- * This "cs" is not working correctly. What is the right one?
- */
-#if 0
+#if 0	/* The scroll region is not working as Vim expects. */
 #  ifdef TERMINFO
     {(int)KS_CS,	"\033[%i%p1%d;%p2%dr"},
 #  else
-     {(int)KS_CS,	"\033[%i%d;%dr"},
+    {(int)KS_CS,	"\033[%i%d;%dr"},
 #  endif
 #endif
     {(int)KS_CL,	"\033[H\033[2J"},
-    {(int)KS_VE,	"\033[9/y\033[12/y\033[=6l"},
-    {(int)KS_VS,	"\033[10/y\033[=1h\033[=2l\033[=6h"},
+    {(int)KS_VE,	"\033[9/y\033[12/y"},
+    {(int)KS_VS,	"\033[10/y\033[=1h\033[=2l"},
+    {(int)KS_TI,	"\033[=6h"},
+    {(int)KS_TE,	"\033[=6l"},
     {(int)KS_SE,	"\033[m"},
     {(int)KS_SO,	"\033[1;7m"},
     {(int)KS_ME,	"\033[m"},
     {(int)KS_MR,	"\033[7m"},
     {(int)KS_MD,	"\033[1m"},
     {(int)KS_UE,	"\033[m"},
+    {(int)KS_CCO,	"8"},			/* allow 8 colors */
+#  ifdef TERMINFO
+    {(int)KS_CAB,	"\033[4%p1%dm"},	/* set background color */
+    {(int)KS_CAF,	"\033[3%p1%dm"},	/* set foreground color */
+#  else
+    {(int)KS_CAB,	"\033[4%dm"},		/* set background color */
+    {(int)KS_CAF,	"\033[3%dm"},		/* set foreground color */
+#  endif
     {(int)KS_US,	"\033[4m"},
     {(int)KS_MS,	"y"},	/* does this really work? */
     {(int)KS_LE,	"\010"},
@@ -3010,18 +3023,28 @@ check_termcode(max_offset)
 		    if (termcodes[i].name[0] == KS_EXTRA)
 			switch (termcodes[i].name[1])
 			{
-			    case KE_XF1: termcodes[i].name[0] = 'k';
-					 termcodes[i].name[1] = '1';
-					 break;
-			    case KE_XF2: termcodes[i].name[0] = 'k';
-					 termcodes[i].name[1] = '2';
-					 break;
-			    case KE_XF3: termcodes[i].name[0] = 'k';
-					 termcodes[i].name[1] = '3';
-					 break;
-			    case KE_XF4: termcodes[i].name[0] = 'k';
-					 termcodes[i].name[1] = '4';
-					 break;
+			    case KE_XF1:
+				termcodes[i].name[0] = KEY2TERMCAP0(K_F1);
+				termcodes[i].name[1] = KEY2TERMCAP1(K_F1);
+				break;
+			    case KE_XF2:
+				termcodes[i].name[0] = KEY2TERMCAP0(K_F2);
+				termcodes[i].name[1] = KEY2TERMCAP1(K_F2);
+				break;
+			    case KE_XF3:
+				termcodes[i].name[0] = KEY2TERMCAP0(K_F3);
+				termcodes[i].name[1] = KEY2TERMCAP1(K_F3);
+				break;
+			    case KE_XF4:
+				termcodes[i].name[0] = KEY2TERMCAP0(K_F4);
+				termcodes[i].name[1] = KEY2TERMCAP1(K_F4);
+				break;
+			    case KE_XEND:
+				termcodes[i].name[0] = KEY2TERMCAP0(K_END);
+				termcodes[i].name[1] = KEY2TERMCAP1(K_END);
+			    case KE_XHOME:
+				termcodes[i].name[0] = KEY2TERMCAP0(K_HOME);
+				termcodes[i].name[1] = KEY2TERMCAP1(K_HOME);
 			}
 
 		    key_name[0] = termcodes[i].name[0];
@@ -3255,12 +3278,21 @@ check_termcode(max_offset)
 	    else
 	    {
 #if defined(UNIX) && defined(HAVE_GETTIMEOFDAY) && defined(HAVE_SYS_TIME_H)
-# ifdef USE_GUI
+# ifdef GPM_MOUSE
+                extern int gpm_flag; /* gpm library variable */
+#  ifdef USE_GUI
 		/*
-		 * Only for Unix, when GUI is not active, we handle
+		 * Only for Unix, when GUI or gpm is not active, we handle
 		 * multi-clicks here.
 		 */
-		if (!gui.in_use)
+                if (gpm_flag == 0 && !gui.in_use)
+#  else
+                if (gpm_flag == 0)
+#  endif
+# else
+#  ifdef USE_GUI
+                if (!gui.in_use)
+#  endif
 # endif
 		{
 		    /*
@@ -3287,7 +3319,7 @@ check_termcode(max_offset)
 		    orig_mouse_row = mouse_row;
 		    orig_topline = curwin->w_topline;
 		}
-# ifdef USE_GUI
+# if defined(USE_GUI) || defined(GPM_MOUSE)
 		else
 		    orig_num_clicks = NUM_MOUSE_CLICKS(mouse_code);
 # endif
@@ -3746,6 +3778,101 @@ show_one_termcode(name, code, printit)
     return len;
 }
 
+/*
+ * Translate an internal mapping/abbreviation representation into the
+ * corresponding external one recognized by :map/:abbrev commands;
+ * respects the current B/k/< settings of 'cpoption'.
+ *
+ * This function is called when expanding mappings/abbreviations on the
+ * command-line, and for building the "Ambiguous mapping..." error messge.
+ *
+ * It uses a growarray to build the translation string since the
+ * latter can be wider than the original description. The caller has to
+ * free the string afterwards.
+ *
+ * Returns NULL when there is a problem.
+ */
+    char_u *
+translate_mapping(str, expmap)
+    char_u	*str;
+    int		expmap;  /* TRUE when expanding mappings on command-line */
+{
+    struct growarray	ga;
+    int			c;
+    int			modifiers;
+    int			cpo_bslash;
+    int			cpo_special;
+    int			cpo_keycode;
+
+    ga_init(&ga);
+    ga.ga_itemsize = 1;
+    ga.ga_growsize = 40;
+
+    cpo_bslash = (vim_strchr(p_cpo, CPO_BSLASH) != NULL);
+    cpo_special = (vim_strchr(p_cpo, CPO_SPECI) != NULL);
+    cpo_keycode = (vim_strchr(p_cpo, CPO_KEYCODE) == NULL);
+
+    for (; *str; ++str)
+    {
+	c = *str;
+	if (c == K_SPECIAL && str[1] && str[2])
+	{
+	    modifiers = 0;
+	    if (str[1] == KS_MODIFIER)
+	    {
+		str++;
+		modifiers = *++str;
+		c = *++str;
+	    }
+	    if (cpo_special && cpo_keycode && c == K_SPECIAL && !modifiers)
+	    {
+		int	i;
+
+		/* try to find special key in termcodes */
+		for (i = 0; i < tc_len; ++i)
+		    if (termcodes[i].name[0] == str[1]
+					    && termcodes[i].name[1] == str[2])
+			break;
+		if (i < tc_len)
+		{
+		    ga_concat(&ga, termcodes[i].code);
+		    str += 2;
+		    continue; /* for (str) */
+		}
+	    }
+	    if (c == K_SPECIAL)
+	    {
+		if (expmap && cpo_special)
+		{
+		    ga_clear(&ga);
+		    return NULL;
+		}
+		c = TO_SPECIAL(str[1], str[2]);
+		if (c == K_ZERO)	/* display <Nul> as ^@ */
+		    c = NUL;
+		str += 2;
+	    }
+	    if (IS_SPECIAL(c) || modifiers)	/* special key */
+	    {
+		if (expmap && cpo_special)
+		{
+		    ga_clear(&ga);
+		    return NULL;
+		}
+		ga_concat(&ga, get_special_key_name(c, modifiers));
+		continue; /* for (str) */
+	    }
+	}
+	if (c == ' ' || c == '\t' || c == Ctrl('J') || c == Ctrl('V')
+	    || (c == '<' && !cpo_special) || (c == '\\' && !cpo_bslash))
+	    ga_append(&ga, cpo_bslash ? Ctrl('V') : '\\');
+	if (c)
+	    ga_append(&ga, c);
+    }
+    ga_append(&ga, NUL);
+    return (char_u *)(ga.ga_data);
+}
+
 #if (defined(WIN32) && !defined(USE_GUI)) || defined(PROTO)
 static char ksme_str[20];
 static char ksmr_str[20];
